/**
 * Copyright (c) 2016 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>
#include <string.h>
#include "nordic_common.h"
#include "app_error.h"
#include "app_uart.h"
#include "ble_db_discovery.h"
#include "app_timer.h"
#include "app_util.h"
#include "bsp_btn_ble.h"
#include "ble.h"
#include "ble_gap.h"
#include "ble_hci.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include "ble_nus_c.h"
#include "nrf_ble_gatt.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_ble_scan.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "nrf.h"
#include "bsp.h"
#include "app_pwm.h"
#include "nrf_drv_gpiote.h"
#include "boards.h"
#include "sdk_errors.h"
#include "nrf_drv_clock.h"
#include "app_util_platform.h"
#include "low_power_pwm.h"
#include "nrf_drv_saadc.h"
#include "nrf_drv_ppi.h"
#include "nrf_drv_timer.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_nvmc.h"
#include "nrf_cli.h"
#include "nrfx_timer.h"
#include "nrf_drv_uart.h"
#include "nrf_cli_uart.h"
#include "nrf_nvmc.h"
#include "nrf_queue.h" 
#include "fds.h"
#include "fds_example.h"
#include "nrf_log_internal.h"

#define NRF_LOG_MODULE_NAME app
#define NRF_CLI_ENABLED 1

#define APP_BLE_CONN_CFG_TAG    1                   /**< Tag that refers to the BLE stack configuration set with @ref sd_ble_cfg_set. The default tag is @ref BLE_CONN_CFG_TAG_DEFAULT. */
#define APP_BLE_OBSERVER_PRIO   3                   /**< BLE observer priority of the application. There is no need to modify this value. */

#define UART_TX_BUF_SIZE        256                 /**< UART TX buffer size. */
#define UART_RX_BUF_SIZE        256                 /**< UART RX buffer size. */

#define NUS_SERVICE_UUID_TYPE   BLE_UUID_TYPE_VENDOR_BEGIN        /**< UUID type for the Nordic UART Service (vendor specific). */

/******************************* DFH *************************/
#define ECHOBACK_BLE_UART_DATA  0                   /**< Echo the UART data that is received over the Nordic UART Service (NUS) back to the sender. */
/******************************* DFH *************************/

#define Board_PS  false             // Change to true for Perfect Shower Valve build, false for nRF52DK development board
#define Board_52DK !Board_PS        // Used for debugging on nRF52DK system

// Pin Definitions for Perfect Shower RevD (MS50SFB) PCBA
#if Board_PS
// nRF52832 GPIO#                 // 52832 Pin# - Port0 # - MS50SFB Pin#
#define STEP_A          28        // pin #40 - P0_28/AIN4 - 6
#define STEP_B          29        // pin #41 - P0_29/AIN5 - 7
#define LED_1           17        // pin #20 - P0.17      - 32 Grn 
#define LED_2           18        // pin #21 - P0.18      - 33 Red  
#define SYS_V           30        // pin #42 - P0.30/AIN6 - 8
#define A_TEMP_OUT      31        // pin #43 - P0.31/AIN7 - 9
#define A_TEMP_IN       2         // pin #4  - P0.02/AIN0 - 16
#define AUDIO_OUT       3         // pin #5  - P0.03/AIN1 - 17
#define MAN_OVERRIDE    11        // pin #14 - P0.11      - 26
#define MAN_RESET       21        // pin #24 - P0.21/RST  - 36
#define TEMP_OVERRIDE   22        // pin #27 - P0.22      - 39
#define D_TEMP          23        // pin #28 - P0.23      - 40
#define ONEWIRE_PIN     23        // def for onewire code
#define TRIP_TEST       24        // pin #29 - P0.24      - 41
#endif

/* Pin Definitions for nRF52DK system */
/* nRF52DK GPIO#                  // 52832 Pin # - Port0.# - DK Hdr */
#if Board_52DK
#define LED_1           17        // pin #20 - P0.17 LED_1 - P3 Grn
#define LED_2           18        // pin #21 - P0.18 LED_2 - P3 Grn
#define STEP_A          28        // pin #40 - P0_28/AIN4  - P2 
#define STEP_B          29        // pin #41 - P0_29/AIN5  - P2
#define SYS_V           30        // pin #42 - P0.30/AIN6  - P2
#define A_TEMP_OUT      31        // pin #43 - P0.31/AIN7  - P2
#define A_TEMP_IN       2         // pin #4  - P0.02/AIN0  - P4
#define AUDIO_OUT       3         // pin #5  - P0.03/AIN1  - P2
#define MAN_OVERRIDE    11        // pin #14 - P0.11       - P3
#define MAN_RESET       21        // pin #24 - P0.21/RST   - P6
#define TEMP_OVERRIDE   22        // pin #27 - P0.22       - P4
#define D_TEMP          23        // pin #28 - P0.23       - P4
#define ONEWIRE_PIN     23        // def for onewire code
#define TRIP_TEST       24        // pin #29 - P0.24       - P4
#endif

// Unused Pins
#define P0_25           25        // pin #37
#define P0_26           26        // pin #38
#define P0_27           27        // pin #39
#define P0_00_XAL1      0         // pin #2
#define P0_01_XAL2      1         // pin #3
#define P0_04_AIN2      4         // pin #6
#define P0_05_AIN3      5         // pin #7
#define P0_06           6         // pin #8
#define P0_07           7         // pin #9
#define P0_08           8         // pin #10
#define P0_09_NFC1      9         // pin #11
#define P0_10_NFC2      10        // pin #12
#define P0_11           11        // pin #14
#define P0_12           12        // pin #15
#define P0_13           13        // pin #16
#define P0_14           14        // pin #17
#define P0_15           15        // pin #18
#define P0_16           16        // pin #19
#define P0_19           19        // pin #22
#define P0_20           20        // pin #23
             
#define LEDS_NUMBER     2
#define LEDS_ACTIVE_STATE 0

#define LEDS_INV_MASK  LEDS_MASK
#define LEDS_LIST {LED_1, LED_2}
#define BSP_LED_0      LED_1
#define BSP_LED_1      LED_2

#define RX_PIN_NUMBER  8
#define TX_PIN_NUMBER  6
#define CTS_PIN_NUMBER 7
#define RTS_PIN_NUMBER 5
#define HWFC           true


/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/


#define ECHOBACK_BLE_UART_DATA  0   /**< Echo the UART data that is received over the Nordic UART Service (NUS) back to the sender. */
#define SETSDATA(A,B,C,D,E,F) \
sdata[0]=A; sdata[1]=B; sdata[2]=C; sdata[3]=D; sdata[4]=E; sdata[5]=F;

#define NRFX_SYSTICK_ENABLED  SYSTICK_ENABLED         // Eric
#define NRFX_TIMER1_ENABLED  TIMER1_ENABLED           // Eric enabled in sdk_config.h as well

uint32_t sys_timer;
uint32_t rc;                                          // return code


/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

BLE_NUS_C_DEF(m_ble_nus_c);                          /**< BLE Nordic UART Service (NUS) client instance. */
NRF_BLE_GATT_DEF(m_gatt);                            /**< GATT module instance. */
BLE_DB_DISCOVERY_DEF(m_db_disc);                     /**< Database discovery module instance. */
NRF_BLE_SCAN_DEF(m_scan);                            /**< Scanning Module instance. */

static uint16_t m_ble_nus_max_data_len = BLE_GATT_ATT_MTU_DEFAULT - OPCODE_LENGTH - HANDLE_LENGTH; /**< Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module. */

/**@brief NUS UUID. */
static ble_uuid_t const m_nus_uuid =
{
    .uuid = BLE_UUID_NUS_SERVICE,
    .type = NUS_SERVICE_UUID_TYPE
};

static char const m_target_periph_name[] = "WLC-01";     /**< Name of the device we try to connect to. This name is searched in the scan report data*/

/**@brief Function for handling asserts in the SoftDevice.
 *
 * @details This function is called in case of an assert in the SoftDevice.
 *
 * @warning This handler is only an example and is not meant for the final product. You need to analyze
 *          how your product is supposed to react in case of assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num     Line number of the failing assert call.
 * @param[in] p_file_name  File name of the failing assert call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}


/**@brief Function for starting scanning. */
static void scan_start(void)
{
    ret_code_t ret;

    ret = nrf_ble_scan_start(&m_scan);
    APP_ERROR_CHECK(ret);

    ret = bsp_indication_set(BSP_INDICATE_SCANNING);
    APP_ERROR_CHECK(ret);
}


/**@brief Function for handling Scanning Module events.
 */
static void scan_evt_handler(scan_evt_t const * p_scan_evt)
{
    ret_code_t err_code;

    switch(p_scan_evt->scan_evt_id)
    {
         case NRF_BLE_SCAN_EVT_CONNECTING_ERROR:
         {
              err_code = p_scan_evt->params.connecting_err.err_code;
              APP_ERROR_CHECK(err_code);
         } break;

         case NRF_BLE_SCAN_EVT_CONNECTED:
         {
              ble_gap_evt_connected_t const * p_connected =
                               p_scan_evt->params.connected.p_connected;
             // Scan is automatically stopped by the connection.
             NRF_LOG_INFO("Connecting to target %02x%02x%02x%02x%02x%02x",
                      p_connected->peer_addr.addr[0],
                      p_connected->peer_addr.addr[1],
                      p_connected->peer_addr.addr[2],
                      p_connected->peer_addr.addr[3],
                      p_connected->peer_addr.addr[4],
                      p_connected->peer_addr.addr[5]
                      );
         } break;

         case NRF_BLE_SCAN_EVT_SCAN_TIMEOUT:
         {
             NRF_LOG_INFO("Scan timed out.");
             scan_start();
         } break;

         default:
             break;
    }
}


/**@brief Function for initializing the scanning and setting the filters.
 */
static void scan_init(void)
{
    ret_code_t          err_code;
    nrf_ble_scan_init_t init_scan;

    memset(&init_scan, 0, sizeof(init_scan));

    init_scan.connect_if_match = true;
    init_scan.conn_cfg_tag     = APP_BLE_CONN_CFG_TAG;

    err_code = nrf_ble_scan_init(&m_scan, &init_scan, scan_evt_handler);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_ble_scan_filter_set(&m_scan, SCAN_NAME_FILTER, &m_target_periph_name);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_ble_scan_filters_enable(&m_scan, NRF_BLE_SCAN_NAME_FILTER, false);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling database discovery events.
 *
 * @details This function is a callback function to handle events from the database discovery module.
 *          Depending on the UUIDs that are discovered, this function forwards the events
 *          to their respective services.
 *
 * @param[in] p_event  Pointer to the database discovery event.
 */
static void db_disc_handler(ble_db_discovery_evt_t * p_evt)
{
    ble_nus_c_on_db_disc_evt(&m_ble_nus_c, p_evt);
}


/**@brief Function for handling characters received by the Nordic UART Service (NUS).
 *
 * @details This function takes a list of characters of length data_len and prints the characters out on UART.
 *          If @ref ECHOBACK_BLE_UART_DATA is set, the data is sent back to sender.
 */
static void ble_nus_chars_received_uart_print(uint8_t * p_data, uint16_t data_len)
{
    ret_code_t ret_val;

    NRF_LOG_DEBUG("Receiving data in received_uart_print.");
    NRF_LOG_HEXDUMP_DEBUG(p_data, data_len);

    for (uint32_t i = 0; i < data_len; i++)
    {
        do
        {
            ret_val = app_uart_put(p_data[i]);
            if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY))
            {
                NRF_LOG_ERROR("app_uart_put failed for index 0x%04x.", i);
                APP_ERROR_CHECK(ret_val);
            }
        } while (ret_val == NRF_ERROR_BUSY);
    }
    if (p_data[data_len-1] == '\r')
    {
        while (app_uart_put('\n') == NRF_ERROR_BUSY);
    }
    if (ECHOBACK_BLE_UART_DATA)
    {
        // Send data back to the peripheral.
        do
        {
            ret_val = ble_nus_c_string_send(&m_ble_nus_c, p_data, data_len);
            if ((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY))
            {
                NRF_LOG_ERROR("Failed sending NUS message. Error 0x%x. ", ret_val);
                APP_ERROR_CHECK(ret_val);
            }
        } while (ret_val == NRF_ERROR_BUSY);
    }
}


/**@brief   Function for handling app_uart events.
 *
 * @details This function receives a single character from the app_uart module and appends it to
 *          a string. The string is sent over BLE when the last character received is a
 *          'new line' '\n' (hex 0x0A) or if the string reaches the maximum data length.
 */
void uart_event_handle(app_uart_evt_t * p_event)
{
    static uint8_t data_array[BLE_NUS_MAX_DATA_LEN];
    static uint16_t index = 0;
    uint32_t ret_val;

    switch (p_event->evt_type)
    {
        /**@snippet [Handling data from UART] */
        case APP_UART_DATA_READY:
            UNUSED_VARIABLE(app_uart_get(&data_array[index]));
            index++;

            if ((data_array[index - 1] == '\n') || (index >= (m_ble_nus_max_data_len)))
            {
                NRF_LOG_DEBUG("Ready to send data over BLE NUS");
                NRF_LOG_HEXDUMP_DEBUG(data_array, index);

                do
                {
                    ret_val = ble_nus_c_string_send(&m_ble_nus_c, data_array, index);
                    if ( (ret_val != NRF_ERROR_INVALID_STATE) && (ret_val != NRF_ERROR_RESOURCES) )
                    {
                        APP_ERROR_CHECK(ret_val);
                    }
                } while (ret_val == NRF_ERROR_RESOURCES);

                index = 0;
            }
            break;

        /**@snippet [Handling data from UART] */
        case APP_UART_COMMUNICATION_ERROR:
            NRF_LOG_ERROR("Communication error occurred while handling UART.");
            APP_ERROR_HANDLER(p_event->data.error_communication);
            break;

        case APP_UART_FIFO_ERROR:
            NRF_LOG_ERROR("Error occurred in FIFO module used by UART.");
            APP_ERROR_HANDLER(p_event->data.error_code);
            break;

        default:
            break;
    }
}

//**********************/
//**********************            ERIK      ******************/
//**********************/

//  ERIC -- please flesh out these dummy routines:

struct blockstruct // block to write into flash.
{
    int desired_temp; // set by user request, used to adjust valves.
    int desired_flow; // same
    int abs_max_temp;

} flashblock;


int get_actual_temp()
{
    //  Eric: please insert code to query the temp sensor.
    //  But for now: return something comfortable.
    return 99;
}

void user_requested_temp(int degF)
{
    flashblock.desired_temp = degF;
    //  Eric: please insert code to adjust the valves.
}

// We don't control flow rate yet so we'll ignore for now

void user_requested_flow(int flow)
{
    flashblock.desired_flow = flow;
    //  Eric: please insert code to adjust the valves.
}


void write_flashblock()
{
    // bunch of NRF stuff.
}

void read_flashblock()
{
    // bunch of NRF stuff.
}

//**********************/
//**********************            ERIK      ******************/
//**********************/

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

void ble_handle_message(char *pdata, int len)
{
    unsigned char sdata[20];
    int i, ndata;

    //  Validate good data.

    if(len <= 5 ||
        pdata[0] != 0xfd || pdata[1] != 0xfc ||
        pdata[2] != len-4)
    {
        NRF_LOG_INFO("Message reject point 341.");
        goto reject;
    }

    uint32_t cs = 0;
    for(i=0; i< (3+pdata[2]); i++)
        cs += pdata[i];
    cs = 0xff & ((~cs) + 1);
    if(cs != pdata[i])
    {
        NRF_LOG_INFO("Message reject checksum, i %d, sb %x, was %x.", i, cs, pdata[i]);
        goto reject;
    }

    //  Analyze message per EPOB0178:

    if(pdata[3] == 0x01 && pdata[2] == 1)
    {
        NRF_LOG_INFO("Remove valve data cmd.");
        //  Replying success.
        SETSDATA(0xfd, 0xfc, 0x02, 0x81, 0x01, 0x00);
        ndata = 5;
        goto sendreply;
    }

    else if(pdata[3] == 0x02 && pdata[2] == 1)
    {
        NRF_LOG_INFO("Set BT work mode cmd.");
        //  Replying success.
        SETSDATA(0xfd, 0xfc, 0x02, 0x82, 0x01, 0x00);
        ndata = 5;
        goto sendreply;
    }

    else if(pdata[3] == 0x03 && pdata[2] == 2)
    {
        NRF_LOG_INFO("Water temp query, scale %d.", pdata[4]);
        //  Replying water temp = 102 degF.
        SETSDATA(0xfd, 0xfc, 0x03, 0x83, 0x00,102);
        ndata = 6;
       goto sendreply;
    }

    else if(pdata[3] == 0x04 && pdata[2] == 3)
    {
        NRF_LOG_INFO("Water temp cmd, scale %d, degs %d.", pdata[4], pdata[5]);
        user_requested_temp(pdata[5]); // assume Fahrenheit.

        //  Replying success.
        SETSDATA(0xfd, 0xfc, 0x02, 0x84, 0x01, 0x00);
        ndata = 5;
        goto sendreply;
    }

    else if(pdata[3] == 0x05 && pdata[2] == 2)
    {
        NRF_LOG_INFO("Water flow cmd, flow %d.", pdata[4]);
        user_requested_flow(pdata[4]);

        //  Replying success.
        SETSDATA(0xfd, 0xfc, 0x02, 0x85, 0x01, 0x00);
        ndata = 5;
        goto sendreply;
    }

    else
    {
        NRF_LOG_INFO("Other command received, %x, %x, %x.", pdata[4], pdata[5], pdata[6]);
        //  Not replying.
        goto reject;
    }


    sendreply:

    //  Compute checksum.

    cs = 0;
    for(i=0; i<ndata; i++)
        cs += sdata[i];
    cs = 0xff & ((~cs) + 1);
    sdata[ndata++] = cs;

    //  Send reply.

    uint32_t ret_val = 21; // arbitrary marker number.
    do
    {
        ret_val = ble_nus_c_string_send(&m_ble_nus_c, sdata, ndata);
        if((ret_val != NRF_SUCCESS) && (ret_val != NRF_ERROR_BUSY))
        {
            NRF_LOG_ERROR("Failed sending NUS message. Error 0x%x. ", ret_val);
            APP_ERROR_CHECK(ret_val);
        }
    } while (ret_val == NRF_ERROR_BUSY);

    reject:
    i=1; // dummy statement.

}

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/


/**@brief Callback handling Nordic UART Service (NUS) client events.
 *
 * @details This function is called to notify the application of NUS client events.
 *
 * @param[in]   p_ble_nus_c   NUS client handle. This identifies the NUS client.
 * @param[in]   p_ble_nus_evt Pointer to the NUS client event.
 */

/**@snippet [Handling events from the ble_nus_c module] */
static void ble_nus_c_evt_handler(ble_nus_c_t * p_ble_nus_c, ble_nus_c_evt_t const * p_ble_nus_evt)
{
    ret_code_t err_code;

    switch (p_ble_nus_evt->evt_type)
    {
        case BLE_NUS_C_EVT_DISCOVERY_COMPLETE:
            NRF_LOG_INFO("Discovery complete.");
            err_code = ble_nus_c_handles_assign(p_ble_nus_c, p_ble_nus_evt->conn_handle, &p_ble_nus_evt->handles);
            APP_ERROR_CHECK(err_code);

            err_code = ble_nus_c_tx_notif_enable(p_ble_nus_c);
            APP_ERROR_CHECK(err_code);
            NRF_LOG_INFO("Connected to device with Nordic UART Service.");
            break;

        case BLE_NUS_C_EVT_NUS_TX_EVT:
            ble_nus_chars_received_uart_print(p_ble_nus_evt->p_data, p_ble_nus_evt->data_len);

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

            NRF_LOG_INFO("NUS TX EVT occurred.");
 
            ble_handle_message(p_ble_nus_evt->p_data, p_ble_nus_evt->data_len);

/******************************* DFH *************************/
/******************************* DFH *************************/
/******************************* DFH *************************/

            break;

        case BLE_NUS_C_EVT_DISCONNECTED:
            NRF_LOG_INFO("Disconnected.");
            scan_start();
            break;
    }
}
/**@snippet [Handling events from the ble_nus_c module] */


/**
 * @brief Function for handling shutdown events.
 *
 * @param[in]   event       Shutdown type.
 */
static bool shutdown_handler(nrf_pwr_mgmt_evt_t event)
{
    ret_code_t err_code;

    err_code = bsp_indication_set(BSP_INDICATE_IDLE);
    APP_ERROR_CHECK(err_code);

    switch (event)
    {
        case NRF_PWR_MGMT_EVT_PREPARE_WAKEUP:
            // Prepare wakeup buttons.
            err_code = bsp_btn_ble_sleep_mode_prepare();
            APP_ERROR_CHECK(err_code);
            break;

        default:
            break;
    }

    return true;
}

NRF_PWR_MGMT_HANDLER_REGISTER(shutdown_handler, APP_SHUTDOWN_HANDLER_PRIORITY);


/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
    ret_code_t            err_code;
    ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            err_code = ble_nus_c_handles_assign(&m_ble_nus_c, p_ble_evt->evt.gap_evt.conn_handle, NULL);
            APP_ERROR_CHECK(err_code);

            err_code = bsp_indication_set(BSP_INDICATE_CONNECTED);
            APP_ERROR_CHECK(err_code);

            // start discovery of services. The NUS Client waits for a discovery result
            err_code = ble_db_discovery_start(&m_db_disc, p_ble_evt->evt.gap_evt.conn_handle);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GAP_EVT_DISCONNECTED:

            NRF_LOG_INFO("Disconnected. conn_handle: 0x%x, reason: 0x%x",
                         p_gap_evt->conn_handle,
                         p_gap_evt->params.disconnected.reason);
            break;

        case BLE_GAP_EVT_TIMEOUT:
            if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
            {
                NRF_LOG_INFO("Connection Request timed out.");
            }
            break;

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            // Pairing not supported.
            err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
            // Accepting parameters requested by peer.
            err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
                                                    &p_gap_evt->params.conn_param_update_request.conn_params);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
        {
            NRF_LOG_DEBUG("PHY update request.");
            ble_gap_phys_t const phys =
            {
                .rx_phys = BLE_GAP_PHY_AUTO,
                .tx_phys = BLE_GAP_PHY_AUTO,
            };
            err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_GATTC_EVT_TIMEOUT:
            // Disconnect on GATT Client timeout event.
            NRF_LOG_DEBUG("GATT Client Timeout.");
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GATTS_EVT_TIMEOUT:
            // Disconnect on GATT Server timeout event.
            NRF_LOG_DEBUG("GATT Server Timeout.");
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break;

        default:
            break;
    }
}


/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    ret_code_t err_code;

    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    // Enable BLE stack.
    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    // Register a handler for BLE events.
    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}


/**@brief Function for handling events from the GATT library. */
void gatt_evt_handler(nrf_ble_gatt_t * p_gatt, nrf_ble_gatt_evt_t const * p_evt)
{
    if (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED)
    {
        NRF_LOG_INFO("ATT MTU exchange completed.");

        m_ble_nus_max_data_len = p_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
        NRF_LOG_INFO("Ble NUS max data length set to 0x%X(%d)", m_ble_nus_max_data_len, m_ble_nus_max_data_len);
    }
}


/**@brief Function for initializing the GATT library. */
void gatt_init(void)
{
    ret_code_t err_code;

    err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_ble_gatt_att_mtu_central_set(&m_gatt, NRF_SDH_BLE_GATT_MAX_MTU_SIZE);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling events from the BSP module.
 *
 * @param[in] event  Event generated by button press.
 */
void bsp_event_handler(bsp_event_t event)
{
    ret_code_t err_code;

    switch (event)
    {
        case BSP_EVENT_SLEEP:
            nrf_pwr_mgmt_shutdown(NRF_PWR_MGMT_SHUTDOWN_GOTO_SYSOFF);
            break;

        case BSP_EVENT_DISCONNECT:
            err_code = sd_ble_gap_disconnect(m_ble_nus_c.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            if (err_code != NRF_ERROR_INVALID_STATE)
            {
                APP_ERROR_CHECK(err_code);
            }
            break;

        default:
            break;
    }
}

/**@brief Function for initializing the UART. */
static void uart_init(void)
{
    ret_code_t err_code;

    app_uart_comm_params_t const comm_params =
    {
        .rx_pin_no    = RX_PIN_NUMBER,
        .tx_pin_no    = TX_PIN_NUMBER,
        .rts_pin_no   = RTS_PIN_NUMBER,
        .cts_pin_no   = CTS_PIN_NUMBER,
        .flow_control = APP_UART_FLOW_CONTROL_DISABLED,
        .use_parity   = false,
        .baud_rate    = UART_BAUDRATE_BAUDRATE_Baud115200
    };

    APP_UART_FIFO_INIT(&comm_params,
                       UART_RX_BUF_SIZE,
                       UART_TX_BUF_SIZE,
                       uart_event_handle,
                       APP_IRQ_PRIORITY_LOWEST,
                       err_code);

    APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the Nordic UART Service (NUS) client. */
static void nus_c_init(void)
{
    ret_code_t       err_code;
    ble_nus_c_init_t init;

    init.evt_handler = ble_nus_c_evt_handler;

    err_code = ble_nus_c_init(&m_ble_nus_c, &init);
    APP_ERROR_CHECK(err_code);
}


/* Function for initializing timer, pins, buttons and leds. */
static void pins_buttons_leds_init(void)
{
    ret_code_t err_code;
    bsp_event_t startup_event;

    err_code = bsp_init(BSP_INIT_LEDS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);

// Set STEP_A and STEP_B pins as standard outputs and set the outputs low
// uint32_t STEP_A, STEP_B;
    nrf_gpio_cfg_output(STEP_A);        // Set pins as outputs
    nrf_gpio_cfg_output(STEP_B);
    nrf_gpio_pin_clear(STEP_A);         // Set pins low.
    nrf_gpio_pin_clear(STEP_B);

// Set up our system timer for general operational use
const nrf_drv_timer_t sys_timer = NRF_DRV_TIMER_INSTANCE(1);      // use timer 1 for non-softdevice timers was 0
//    APP_TIMER_DEF(sys_timer);                 // Creates timer id for our program.

}


/**@brief Function for initializing the timer. */
static void timer_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for initializing the nrf log module. */
static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}


/**@brief Function for initializing power management.
 */
static void power_management_init(void)
{
    ret_code_t err_code;
    err_code = nrf_pwr_mgmt_init();
    APP_ERROR_CHECK(err_code);
}


/** @brief Function for initializing the database discovery module. */
static void db_discovery_init(void)
{
    ret_code_t err_code = ble_db_discovery_init(db_disc_handler);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling the idle state (main loop).
 *
 * @details Handles any pending log operations, then sleeps until the next event occurs.
 */
static void idle_state_handle(void)
{
    if (NRF_LOG_PROCESS() == false)
    {
        nrf_pwr_mgmt_run();
    }
}


int main(void)
{
    // Initialize.
    log_init();
    timer_init();
    uart_init();
    pins_buttons_leds_init();                     // Eric - modified

    application_timers_start();                   // Eric - added
    uint32_t err_code = NRF_LOG_INIT(NULL);       // Eric - added
    APP_ERROR_CHECK(err_code);                    // Eric - added
    NRF_LOG_DEFAULT_BACKENDS_INIT();              // Eric - added
    saadc_init();                                 // Eric - added
    saadc_sampling_event_init();                  // Eric - added

    db_discovery_init();
    power_management_init();
    ble_stack_init();
    gatt_init();
    nus_c_init();
    scan_init();

    // Start execution.
    printf("BLE UART central example started.\r\n");
    NRF_LOG_INFO("BLE UART central example started.");
    scan_start();

    // Enter main loop.
    for (;;)
    {
        idle_state_handle();
    }
}

// *************************************************************************************
/* **************** End of Nordic ble_app_uart_c Example Code **************************
// *************************************************************************************

// *************************************************************************************
//       Nordic DS18B20 Temperature sensor interface (FINISHED)
// *************************************************************************************
*/
//  Structure holding a 1-wire serial number. 
typedef struct sBSPACMonewireSerialNumber
{
	// The serial number in order MSB to LSB 
	uint8_t id[6];
} sBSPACMonewireSerialNumber;

enum
{
	BSPACM_ONEWIRE_CMD_READ_ROM = 0x33,                     // Read 64-bit ROM code without using search procedure 
	BSPACM_ONEWIRE_CMD_SKIP_ROM = 0xcc,			// Skip ROM sends the following command to all bus devices 
	BSPACM_ONEWIRE_CMD_READ_POWER_SUPPLY = 0xb4,		// Tw]ƬOHͨѹq٬O~ѹq
	BSPACM_ONEWIRE_CMD_RECALL_EE = 0xb8,			// Store data from EEPROM into RAM 
	BSPACM_ONEWIRE_CMD_READ_SCRATCHPAD = 0xbe,		// Read the RAM area. 
	BSPACM_ONEWIRE_CMD_CONVERT_T = 0x44,			// Initiate a temperature conversion.
															
};

/** Define protocol state times in microseconds.
 *
 * @note Since all these times are far less than any sane watchdog
 * interval, and the timing can be important, BSPACM_CORE_DELAY_CYCLES
 * is not used in this module.  */

enum
{
	OWT_RSTL_us = 480,				// Minimum time to hold bus low to ensure reset
	OWT_RSTH_us = 480,				// Time to wait for presence detection after reset to quiesce 
	OWT_PDHIGH_us = 60,				// Delay before presence pulse is known to be valid (15us..60us) 
	OWT_LOW0_us = 60,				// Minimum time to hold bus low when writing a zero 
	OWT_LOW1_us = 1,				// Minimum time to hold bus low when writing a one 
	OWT_REC_us = 1,					// Recovery delay between write/read transaction cycles 
	OWT_INT_us = 1,					// Time to hold bus low when initiating a read slot 
	OWT_RDV_us = 15 - OWT_INT_us,			// Point at which read value should be sampled 
	OWT_SLOT_us = 60,				// Minimum duration of a read or write slot 
};

volatile bool cc0_timeout;

void TIMER0_IRQHandler ()                  // Set up timer 1 for this?
{
  if (NRF_TIMER0->EVENTS_COMPARE[0]) {     // Set up timer 1 for this?
    cc0_timeout = true;
    NRF_TIMER0->EVENTS_COMPARE[0] = 0;
  }
}

void delay_us (uint32_t count_us)
{

  /* Don't muck with timers if the duration is so short we might miss
   * the compare value wrap. */

  if (2 >= count_us) {
    nrf_delay_us(count_us);
    return;
  }
  cc0_timeout = false;
  NRF_TIMER0->TASKS_CAPTURE[0] = 1;
  NRF_TIMER0->EVENTS_COMPARE[0] = 0;
  NRF_TIMER0->INTENSET = TIMER_INTENSET_COMPARE0_Msk;
  NRF_TIMER0->CC[0] += count_us;
  while (! cc0_timeout) {
    __WFE();
  }
  NRF_TIMER0->INTENCLR = TIMER_INTENCLR_COMPARE0_Msk;
  return;
}

#ifndef ONEWIRE_PIN
#define ONEWIRE_PIN 23                  // P0.23 Dallas D_Temp input
#endif // ONEWIRE_PIN 
#define ONEWIRE_BIT (1ULL << ONEWIRE_PIN)

static void dirset (void)                   // Set's pin as output.
{
	NRF_GPIO->DIRSET = ONEWIRE_BIT;     
}

void dirclr (void)                          // Set's pin as input.
{
	NRF_GPIO->PIN_CNF[ONEWIRE_PIN] = 0
		| (GPIO_PIN_CNF_SENSE_Disabled << GPIO_PIN_CNF_SENSE_Pos)
		| (GPIO_PIN_CNF_DRIVE_S0S1 << GPIO_PIN_CNF_DRIVE_Pos)
		| (GPIO_PIN_CNF_PULL_Disabled << GPIO_PIN_CNF_PULL_Pos)
		| (GPIO_PIN_CNF_INPUT_Connect << GPIO_PIN_CNF_INPUT_Pos)
		| (GPIO_PIN_CNF_DIR_Input << GPIO_PIN_CNF_DIR_Pos);
}

int iBSPACMonewireReset_ni (void)
{
	int present;

#if 1
	// Non-standard: Hold bus high for OWT_RESET_us.	This provides
	// enough parasitic power for the device to signal presence.
	// Without this, effective RSTL duration may exceed the maximum
	// 960us before device reset. 
	dirset();
	NRF_GPIO->OUTSET = ONEWIRE_BIT;
	nrf_delay_us(OWT_RSTH_us);

	// Hold bus low for OWT_RESET_us 
	NRF_GPIO->OUTCLR = ONEWIRE_BIT;
	nrf_delay_us(OWT_RSTL_us);

	// Release bus and switch to input until presence pulse should be
	// visible. 
	dirclr();
	nrf_delay_us(OWT_PDHIGH_us);

	// Record presence if bus is low (DS182x is holding it there) 
	present = !(NRF_GPIO->IN & ONEWIRE_BIT);

	// Wait for reset cycle to complete 
	nrf_delay_us(OWT_RSTH_us - OWT_PDHIGH_us);
#else
	// Non-standard: Hold bus high for OWT_RESET_us.	This provides
	// enough parasitic power for the device to signal presence.
	// Without this, effective RSTL duration may exceed the maximum
	// 960us before device reset. 
	nrf_gpio_cfg_output(ONEWIRE_PIN);
	nrf_gpio_pin_set(ONEWIRE_PIN);
	nrf_delay_us(OWT_RSTH_us);

	// Hold bus low for OWT_RESET_us
	nrf_gpio_pin_clear(ONEWIRE_PIN);
	nrf_delay_us(OWT_RSTL_us);

	nrf_gpio_cfg_input(ONEWIRE_PIN, NRF_GPIO_PIN_NOPULL);			// `uäJAsb߽ӥi
	nrf_delay_us(OWT_PDHIGH_us);

	// Record presence if bus is low (DS182x is holding it there)
	present = !(nrf_gpio_pin_read(ONEWIRE_PIN));
#endif
	// Wait for reset cycle to complete
	nrf_delay_us(OWT_RSTH_us - OWT_PDHIGH_us);

	return present;
}

void vBSPACMonewireShutdown_ni (void)
{
	NRF_GPIO->OUTCLR = ONEWIRE_BIT;
	dirclr();
}

void vBSPACMonewireWriteByte_ni (int byte)
{
	int bp;

	for (bp = 0; bp < 8; ++bp)
	{
		NRF_GPIO->OUTCLR = ONEWIRE_BIT;
		dirset();
		if (byte & 0x01) {
			nrf_delay_us(OWT_LOW1_us);
			dirclr();
			nrf_delay_us(OWT_SLOT_us - OWT_LOW1_us + OWT_REC_us);
		} else {
			nrf_delay_us(OWT_LOW0_us);
			dirclr();
			nrf_delay_us(OWT_SLOT_us - OWT_LOW0_us + OWT_REC_us);
		}
		byte >>= 1;
	}
}

int iBSPACMonewireReadBit_ni (void)
{
	int rv;

	NRF_GPIO->OUTCLR = ONEWIRE_BIT;
	dirset();
	nrf_delay_us(OWT_INT_us);
	dirclr();
	nrf_delay_us(OWT_RDV_us);
	//abc	vBSPACMledSet(0, 1);
	rv = !!(NRF_GPIO->IN & ONEWIRE_BIT);
	//abc	vBSPACMledSet(0, 0);
	nrf_delay_us(OWT_SLOT_us - OWT_RDV_us - OWT_INT_us + OWT_REC_us);
	return rv;
}

int iBSPACMonewireReadByte_ni (void)
{
	int byte = 0;
	int bit = 1;

	do {
		if (iBSPACMonewireReadBit_ni()) {
			byte |= bit;
		}
		bit <<= 1;
	} while (0x100 != bit);
	return byte;
}

int iBSPACMonewireComputeCRC(const unsigned char * romp, int len)
{
	static const unsigned char OW_CRC_POLY = 0x8c;
	unsigned char crc = 0;

	while (0 < len--) {
		int bi;

		crc ^= *romp++;
		for (bi = 0; bi < 8; ++bi) {
			if (crc & 1) {
				crc = (crc >> 1) ^ OW_CRC_POLY;
			} else {
				crc >>= 1;
			}
		}
	}
	return crc;
}

int iBSPACMonewireReadSerialNumber (sBSPACMonewireSerialNumber * snp)
{
	uint8_t rom[8];
	int i;
	int rv = -1;

	do {
		if (! iBSPACMonewireReset_ni()) {
			break;
		}
		vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_READ_ROM);
		for (i = 0; i < sizeof(rom); ++i) {
			rom[i] = iBSPACMonewireReadByte_ni();
		}
		rv = 0;
	} while (0);

	if (0 == rv) {
		if (0 != iBSPACMonewireComputeCRC(rom, sizeof(rom))) {
			rv = -1;
		} else {
			for (i = 0; i < sizeof(snp->id); ++i) {
				snp->id[i] = rom[sizeof(rom) - 2 - i];
			}
		}
	}
	return rv;
}

int iBSPACMonewireRequestTemperature_ni (void)
{
	if (! iBSPACMonewireReset_ni()) {
		return -1;
	}
	vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
	vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_CONVERT_T);
	return 0;
}

int iBSPACMonewireReadPowerSupply (void)
{
	int rv = -1;

	if (iBSPACMonewireReset_ni())
	{
		vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
		vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_READ_POWER_SUPPLY);
		rv = iBSPACMonewireReadBit_ni();
	}
	return rv;
}

int iBSPACMonewireReadTemperature_ni (int * temp_xCel)
{
	int t;

	if (! iBSPACMonewireReset_ni()) {
		return -1;
	}
	vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_SKIP_ROM);
	vBSPACMonewireWriteByte_ni(BSPACM_ONEWIRE_CMD_READ_SCRATCHPAD);
	t = iBSPACMonewireReadByte_ni();
	t |= (iBSPACMonewireReadByte_ni() << 8);
	*temp_xCel = t;
	return 0;
}

int check_temp_probe(void)
{
        if (! iBSPACMonewireReset_ni()) {
	//printf("ERR: No DS18B20 present on P0.%u\n", ONEWIRE_PIN);
	NRF_LOG_INFO("ERR: No DS18B20 present on P0.%u\n", ONEWIRE_PIN);
//	break;
	}

	static  char * const supply_type[] = { "parasitic", "external" };

	int external_power = iBSPACMonewireReadPowerSupply();
	//	printf("Power supply: %s\r\n", supply_type[external_power]);
	NRF_LOG_INFO("Power supply: %s \r\n", nrf_log_push(supply_type[external_power]));
	if (0 > external_power) {
            NRF_LOG_INFO("ERROR: Device not present?\r\n");
//            break;
	}

	sBSPACMonewireSerialNumber serial;
	int rc = iBSPACMonewireReadSerialNumber(&serial);
	//printf("Serial got %d: \r\n", rc);
	NRF_LOG_INFO("Serial got %d: \r\n", rc);
	//abc	vBSPACMconsoleDisplayOctets(serial.id, sizeof(serial.id));
	//abc	putchar('\n');

  return rc;
  dirset();
  NRF_GPIO->OUTCLR = ONEWIRE_BIT;
        
}

int get_temp(void)
{
    while (0 == iBSPACMonewireRequestTemperature_ni())
    {
      int t_raw;
      nrf_delay_ms(600UL); //600UL
        while (! iBSPACMonewireReadBit_ni()) {
        }
        rc = iBSPACMonewireReadTemperature_ni(&t_raw);
        int t_dCel = (10 * t_raw) / 16;                     // perform t_raw to F & C conversions from calling code
        int t_dFahr = 320 + (9 * t_dCel) / 5;
			
        //printf("Got %d dCel, %d d[Fahr]\r\n", t_dCel, t_dFahr);
      NRF_LOG_INFO("Got %d dCel, %d d[Fahr]\n", t_dCel, t_dFahr);
    }
  return rc;         // perform t_raw to F & C conversions from calling code
}
/******************************************************************************/
//            End of Nordic DS18B20 Temperature sensor module
// ****************************************************************************
/*********************************************************************
*                   Stepper Motor Control (FINISHED)
**********************************************************************/

// Stepper Motor Step sequiences for STEP_A & STEP_B GPIOs
// #define STEP_FORWARD[4] {00, 10, 11, 01}      // 0,2,3,1
// #define STEP_BACKWARD[4] {00, 01, 11, 10}      // 0,1,3,2

uint32_t step_pin, stepper_A, stepper_B;              // temporary value of output pin STEP_A & STEP_B    

// Step Forward Direction
void step_forward(uint32_t num_steps)
{
    stepper_A = nrf_gpio_pin_out_read(STEP_A);   // Read the current state of STEP_A and STEP_B pin levels
    stepper_B = nrf_gpio_pin_out_read(STEP_B);

    if (stepper_A && stepper_B || !stepper_A && !stepper_B) {  // Determine which drive pin (STEP_A or STEP_B) needs to be
            step_pin = STEP_B;                                // changed first in order to move in the forward direction.
    } else {
            step_pin = STEP_A;
    }
          // Now Step the requested number of steps
    do {
       nrf_gpio_pin_toggle(step_pin);     // First step in the forward direction
         if (step_pin = STEP_B) {
                step_pin = STEP_A;
         } else {
                step_pin = STEP_B;
         }
       num_steps--;  
       nrf_delay_ms(20);   // 20mS delay. Change this to timer ticks
    }while (num_steps != 0);
}

// Setp Backwards direction
void step_backward(uint32_t num_steps)
{
    stepper_A = nrf_gpio_pin_out_read(STEP_A);   // Read the current state of STEP_A and STEP_B pin levels
    stepper_B = nrf_gpio_pin_out_read(STEP_B);

    if (stepper_A && stepper_B || !stepper_A && !stepper_B){  // Determine which drive pin (STEP_A or STEP_B) needs to be
            step_pin = STEP_A;                                // changed first in order to move in the backward direction.
    } else {
            step_pin = STEP_B;
    }
          // Now Step the requested number of steps
    do {
       nrf_gpio_pin_toggle(step_pin);     // First step in the backward direction
        if (step_pin = STEP_B) {
                step_pin = STEP_A;
         } else {
                step_pin = STEP_B;
         }
       num_steps--;  
       nrf_delay_ms(20);                     // 20mS delay. Change this to timer ticks
    }while (num_steps != 0);
}

/*********************************************************************
*                    End of Stepper Motor Control
**********************************************************************/
/**********************************************************************
************* Eric's Original Arduino Code & New Nordic Code  *********
***********************************************************************
int32_t volatile WG_TEMP;            // Current WaterGeni water temperature
int32_t volatile IPOD_TEMP = 83;     // Start with a reasonable value in case 
                                     //   there is a flash read problem.
int32_t volatile SAVED_TEMP = 107;   // Last temperature saved to flash. Start with 95 F.

// #define  MY_FLASH_PAGE  251          //   A flash page is 1K in length, so page 251 starts
                                     //   at address 251 * 1024 = 257024 = 3EC00 hex
// flash_data *flash = (flash_data*)ADDRESS_OF_PAGE(MY_FLASH_PAGE);      // pointer to flash
// int rc;                              // Return code for flash routine
const uint8_t STEPS_PER_REVOLUTION = 200; // 1.8 degrees per step - Motor design constant
byte i;                              // Dallas temperature sensor variables and data structures.
byte present = 0;                    
byte type_s;
byte data[12];
byte addr[8];     
int32_t CELSIUS, FAHRENHEIT;
int STEP_COUNT = 0;                  // Number of steps the motor has taken.
int lOOP_COUNT = 0;                  // Count the number of times through the main loop 
bool CLOCKWISE_HEAT = true;          // To start, assume that water gets hotter with clockwise.
                                    //    This will get changed by the initialization code and the
                                    //    final value stored in flash for reading during power up.
                                    //    Need to implement flash save!
bool purgingValve = false;          // True only when we're purging the hot water.
int32_t trendTemp1, trendTemp2 = 0; // Temporary temperature value to determine the direction
                                    //    of the changing water temperature, i.e. hotter or colder                        
byte tempReadCount = 0;             // Keeps track of the number of sequential temperature
                                    //    readings for temperature trending info.
bool trendingUp = false;            // which way is the water temperature moving, up or down?
int trendAmount = 0;                // How fast is the water changing temperature?
int32_t absoluteMaxTemp = 119;      // Default = 115 F. This is programmable by the user 
                                    //    interface. This value will be saved in flash. Hardware 
                                    //    safety temperature circuitry set to 120 F. 
bool initializationDone = true;     // When the valve is first installed, the installation code  
                                    //    will set this to "true".  This way we know all the system 
                                    //    variables are configured properly
bool volatile installMode = false;           // First time installation(& reset)mode.
bool volatile userInterfaceConnect = false;  // Set to true by the BLE.connect function.
bool volatile manualMode = false;            // Need to know when the user is controlling things.
bool volatile programMode = false;           // Used to know when we are recording a user programed profile.
bool volatile programRunMode = false;        // Used to know when we are running a pre-programmed profile.
uint8_t bitResolution = (0x5f);     // DS18B20 - 11 bits to speed up temp conversion, don't need full accuracy.
                                    // 5 = 0101 R1,R0 = 10 = 11 bit reslution.
unsigned char cfgRegValue;          // Value stored in the ds18b20 confgiuration register.     
bool volatile resetWG = false;      // If resetWG = true then set initializationDone and installMOde
                                    //   to false so we can reset the valve and perform 
                                    //   installation agian. 
bool clockwise1, clockwise2;        // Used to determine the valve rotation direction.
uint8_t stepperSpeed = 50;          // Number of steps per second of stepper motor.
uint8_t littleStep = 1;             // Number of stepper motor steps (1.8 degrees) to rotate.
uint8_t bigStep = 3;
uint8_t purgeStep = 5;
uint8_t bigpurgeStep = 10;
uint8_t fullOffStep = 110;           // The valve turns 90 from off to full hot.  In case stepper
uint8_t fullOnStep = 110;            //    misses a few steps go a few extra.  Not totally
                                     //    necessary with the reed switches. 
uint32_t userMaxTemp = 107;          // Maximum allowed temperature programmed by user via the U.I.
bool updateFlash = false;            // This flag is used to know when to update the falsh with new values.
uint8_t programBuff1 [2][25] = {{95},{0}}; // Define a program buffer with 25 temperature change values
                                     //    and 25 time durations in seconds(one per temperature setting.)
                                     //    Start out with 95 degrees and unlimited time (0 sec.).
uint8_t col;                         // Storage array Column pointers
int32_t progTemp = 0;                // Temporary storage for user programming temperature profile.
uint32_t runTimeCounter = 0;         // Used for user profile programming

// Controller pin assignments (Arduino version) 
/*
const int system_Voltage = 5;        // GPIO 5 is our reset pin and system voltage monitoring input.
const int manual_Operation = 1;      // GPIO 1 for the manual/automatic operation - handle position.
const int analog_Temp_Out = 0;       // GPIO 0 is analog temperature output to LED bar.
int systemVoltage = 0;               // value read from the GPIO 5 (reset_AnalogInput) - power monitor circuit
int outputTempToDisplay = 0;         // value output to the PWM (analog out) - to LED bar

/**********************************
 * Flash Memory Storage Structure *
 **********************************
*flash[1] = saved_Temp            *
*flash[2] = absoluteMaxTemp       *
*flash[3] = userMaxTemp           *
*flash[4] = clockwiseHeat         *
*flash[5] = installMode           *
*flash[6] = InitializationDone    *
*flash[7] = bitResolution         *
*flash[8] =                       *
*flash[9] = 
*flash[10] = 
***********************************/
/*
// Define the flash structure:
struct flash_data
{
int32_t  f_saved_Temp;
int32_t  f_absoluteMaxTemp;
int32_t  f_userMaxTemp;
bool  f_clockwiseHeat;
bool  f_installMode;
bool  f_initializationDone;
int8_t  f_bitResolution;
};

// Create and fill the structure.
struct flash_data flash_data_values = {saved_Temp, absoluteMaxTemp, userMaxTemp,
clockwiseHeat, installMode, initializationDone, bitResolution};
*/

/*****************  Manual Factory Reset  ********************
//  Read Reset pin GPIO 5 to see if we need to perform a system reset
    int pinState = digitalRead(system_Voltage);  // Need to change to digital input vs analog?
  if (pinState == 0){                            // Set all of the flags to reset system.
    initializationDone = false;
    installMode = true;
  }


  /*************************************************************************
   *                  VALVE ROTATION DIRECTION
   *  This code figures out which direction of rotation makes the water
   *  hotter and colder.  The valve could be plumbed backwards and different
   *  manufacturers have different designs or the stepper motor could be
   *  installed in the WaterGeni upside down.
   *  NOTE: This code is only run during initial installation or after a
   *  reset.  The "bool clockwiseHeat = true/false" is stored in flash for 
   *  subsequent reference during power up so the valve knows which way to turn.
   *************************************************************************

  Serial.println("Waiting for User Interface to connect");
  Serial.println();
  while (!userInterfaceConnect);           // Wait for BLE connection so we know the 
                                           // user interface is in communications with us.
  Serial.println("User Interface Connected!");
  
  while (userInterfaceConnect && installMode){ // Only runs during Installation.

  Serial.println("Entering Installation Mode");
//  RFduinoBLE.send = (myarray, 14);           // Send "Configuring..." to user interface.
    delay(2000);                               // Wait 2 seconds to read message.

//  RFduinoBLE.send = (myarray, 10);           // Send "Purging..." to user interface.
    // Purge valve till water temperature stabilizes. Wait 10 seconds to start.
    for (int i = 10; i >= 1; i--){
      Serial.printf("Purging Hot Water Line... Wait %d seconds\n",i);
      delay(1000);  
//  RFduinoBLE.send = (myarray, 10);           // Send "Purging..." to user interface.
     }

//******** Test 1 **********
    // Wait for water temperature to be stable for three seconds
//  RFduinoBLE.send = (myarray, 8);           // Send "Testing." to user interface.
    while ((WG_Temp != trendTemp1) && (WG_Temp != trendTemp2)){ 
      GetWGTemperature();                     // Read water temperature    
      Serial.printf("Test 1: Sending first temp to iPod = %f F\n",WG_Temp); 
      RFduinoBLE.sendByte(WG_Temp);           // send the temperature to the iPad
      trendTemp2 = trendTemp1;                // Start collecting trend data.  
      trendTemp1 = WG_Temp;          
      delay(1000);     
    }
int32_t testTemp1 = WG_Temp;                   // save this temperature for comparison later.

//******** Test 2 **********
    // Rotate valve clockwise and wait for it to be stable for three seconds again.
//  RFduinoBLE.send = (myarray, 9);           // Send "Testing.." to user interface.
    trendTemp1 = 0;                    // Clear one variable just in case. (shouldn't matter.)
    myStepper.step(10);                // Rotate valve 18 degrees "clockwise."     
    while ((WG_Temp != trendTemp1) && (WG_Temp != trendTemp2)){ // Wait for water to stabize
      GetWGTemperature();                     // Read water temperature    
      Serial.printf("Test 2: Sending second temp to iPod = %f F\n",WG_Temp); 
      RFduinoBLE.sendByte(WG_Temp);           // send the temperature to the iPad
      trendTemp2 = trendTemp1;                // Start collecting trend data.  
      trendTemp1 = WG_Temp;          
      delay(1000);     
    }
int32_t testTemp2 = WG_Temp;           // save this temperature for comparison later.

//******** Test 3 **********
    // Rotate valve clockwise and wait for it to be stable for three seconds again.
//  RFduinoBLE.send = (myarray, 10);           // Send "Testing..." to user interface.
    trendTemp1 = 0;                    // Clear one variable just in case. (shouldn't matter.)
    myStepper.step(10);                // Rotate valve 18 degrees "clockwise."     
    while ((WG_Temp != trendTemp1) && (WG_Temp != trendTemp2)){ // Wait for water to stabize
      GetWGTemperature();                     // Read water temperature    
      Serial.printf("Test 3: Sending third temp to iPod = %f F\n",WG_Temp); 
      RFduinoBLE.sendByte(WG_Temp);           // send the temperature to the iPad
      trendTemp2 = trendTemp1;                // Start collecting trend data.  
      trendTemp1 = WG_Temp;          
      delay(1000);     
    }
int32_t testTemp3 = WG_Temp;           // save this temperature for comparison later.

    // Did we get hotter or colder?
    if ((testTemp3 > testTemp2) && (testTemp2 > testTemp1)){
      clockwise1 = true; 
    }else{
      clockwise1 = false;
    }
    // Lets go the opposite direction to make sure
   
//******** Test 4 **********
    // Wait for water temperature to be stable for three seconds
//  RFduinoBLE.send = (myarray, 11);           // Send "Testing...." to user interface.
    trendTemp1 = 0;                    // Clear one variable just in case. (shouldn't matter.)
    myStepper.step(-10);                // Rotate valve 18 degrees "counterclockwise."     
    while ((WG_Temp != trendTemp1) && (WG_Temp != trendTemp2)){ 
      GetWGTemperature();                     // Read water temperature    
      Serial.printf("Test 4: Sending first temp to iPod = %f F\n",WG_Temp); 
      RFduinoBLE.sendByte(WG_Temp);           // send the temperature to the iPad
      trendTemp2 = trendTemp1;                // Start collecting trend data.  
      trendTemp1 = WG_Temp;          
      delay(1000);     
    }
     testTemp2 = WG_Temp;                   // save this temperature for comparison later.

//******** Test 5 **********
    // Rotate valve clockwise and wait for it to be stable for three seconds again.
//  RFduinoBLE.send = (myarray, 12);           // Send "Testing....." to user interface.
    trendTemp1 = 0;                    // Clear one variable just in case. (shouldn't matter.)
    myStepper.step(-10);                // Rotate valve 18 degrees "counterclockwise."     
    while ((WG_Temp != trendTemp1) && (WG_Temp != trendTemp2)){ // Wait for water to stabize
      GetWGTemperature();                     // Read water temperature    
      Serial.printf("Test 5: Sending second temp to iPod = %f F\n",WG_Temp); 
      RFduinoBLE.sendByte(WG_Temp);           // send the temperature to the iPad
      trendTemp2 = trendTemp1;                // Start collecting trend data.  
      trendTemp1 = WG_Temp;          
      delay(1000);     
    }
    testTemp1 = WG_Temp;           // save this temperature for comparison later.

    // Did we get hotter or colder?  Compare with first test
    if ((testTemp3 > testTemp2) && (testTemp2 > testTemp1)){
      clockwise2 = true;              // We got colder
    }else{
      clockwise2 = false;
    }
    if ((clockwise1 = true) && (clockwise2 = true)){
      clockwiseHeat = true;                  // Clockwise heats the water!
      Serial.printf("*** Initialization PASSED ***\n"); 
      Serial.printf("clockwiseHeat = %b\n", clockwiseHeat); 
//    RFduinoBLE.send = (myarray, 16);     // Send "*** FINISHED ***" to user interface.
      installMode = false;                   // 
      initializationDone = true;  
    }else 
      if ((clockwise1 = false) && (clockwise2 = false)){
        clockwiseHeat = false;              // Counterclockwise heats the water!
        Serial.printf("*** Initialization PASSED ***\n"); 
        Serial.printf("clockwiseHeat = %b\n", clockwiseHeat); 
//      RFduinoBLE.send = (myarray, 16);     // Send "*** FINISHED ***" to user interface.
        installMode = false;                // 
        initializationDone = true;  
    }else {
//      if (((clockwise1 = true) && (clockwise2 = false)) || ((clockwise1 = false) && (clockwise2 = true)))
        Serial.printf("*** Initialization FAILED ***\n"); 
//      RFduinoBLE.send = (myarray, 13);     // Send "*** ERROR ***" to user interface.
        installMode = true;                    // Stay in install mode till it works.
        initializationDone = false; 
//        break;
    }  
// NEED TO WRITE ALL THE CONFIGURATION SETTINGS TO FLASH BEFORE WE TURN OFF!!!
// clockwiseHeat, installMode, initializationDone, etc...
   SaveToFlash();   

// If there was an error during the clockwiseHeat direction testing the installMode and 
// installationDone flags would not have been set so the setup() function will run again
// after the water is manually restarted.  If there was no error, setup() function will 
// run again.

    // Turn off the valve to restart.
      Serial.printf("Turning off water\n"); 
//      RFduinoBLE.send = (myarray, 17);     // Send "Turning off water" to user interface.
   if (clockwiseHeat) {      
      myStepper.step(-fullOffStep);                   // turn valve off     
     }
     else {
      myStepper.step(fullOffStep);
     } 
  }
 
  /*************************************************************************
   *                  Bluetooth LE Transmit Power Level Setting
   *  Need code to adjust the Bluetooth radio power level so that we get
   *  reliable communications between the valve and the user controller.
   *  Since the valve will have a decorative escutcheon plate it is likely 
   *  that it will interfear with the RF signal. Programmable between -20dBm
   *  and +4dBm in 4dBm increments, i.e. -20,-16,-12,-8,-4,0,+4.
   *************************************************************************/
   
  // CODE HERE...(must work along with the user interface unit.)


// *************************   GET SAVED TEMPERATURE  *******************************
  //  READ THE SAVED FLASH VARIABLES TO INITIALIZE THE SYSTEM
//  saved_Temp = flash->f_saved_Temp;
/*
  Read_Flash(flash);
  
  Serial.printf("Temperature initialized to: %d F\n",saved_Temp);
  Serial.printf("clockwiseHeat is: %d\n",clockwiseHeat);
  
  // Now make this the initial water temperature to start
  iPod_Temp = saved_Temp;

  ReadSystemPower();

  Serial.println("*************************************");
  Serial.println("          Entering Main Loop ");
  Serial.println("*************************************");
  Serial.println();
  Serial.println("Waiting for User Interface to connect");

} /********************  END OF SETUP  **********************

/********************************************************************************
*********************************************************************************
                      MAIN LOOP BEGINS HERE
*********************************************************************************
*********************************************************************************
void loop() {
  while (!userInterfaceConnect);          // Wait for BLE connection so we know the 
                                          // user interface is communicating with us.
  Serial.printf("loopCount = %d\n",loopCount); // Used for debug purposes

/**********************  STATR LOOP TIMING TEST SECTION  **************************
//  Test to see how fast we go through the loop (0.395sec w/ monitor running)
bool volatile state;                      // 0.334 sec. without monitor running.
int8_t x;
  x = loopCount % 100;
  Serial.printf("x = %d, and state = %b\n",x,state);
  if ((x == 0)&&(state == true)){
     digitalWrite(4, 0);                  // Toggle the blue LED every 100 loop passes.
     state = false;
  }else {
    if ((x == 0)&&(state == false)){
      Serial.println("Got here");
     digitalWrite(4, 1);    
     state = true;
     Serial.printf("state = %b\n",state);
    }
  }
***********************  END OF LOOP TIMING TEST SECTION  ************************

  ReadSystemPower();                // We need to monitor system power for Flash writes...
                                    //   Probably want to do this in the SaveToFlash routine 
                                    //   to make sure we have enough power to do the write.

  // Get the DESIRED water temperature from iPod
  Serial.printf("iPod_Temp = %d F\n", iPod_Temp);

  GetWGTemperature();  
  Serial.printf("WG Temp  = %d F\n",WG_Temp);

/*******************************  WARNING/DANGER  ***********************************
 *   Make sure that the WG_Temp can never be higher than the absoluteMaxTemp variable
 ************************************************************************************
  if (absoluteMaxTemp > 120) {
    absoluteMaxTemp = 120;
    Serial.println("WARNING, MAX temperature exceeds safe limit");
    Serial.printf("MAX Temperature limited to %d F",absoluteMaxTemp); 
  }
  if (WG_Temp > absoluteMaxTemp) {
    Serial.println("WARNING, WARNING temperature exceeds safe limit");
    Serial.printf("Temperature limited to %d F",absoluteMaxTemp); 
    WG_Temp = absoluteMaxTemp;
  }
    // send the temperature to the iPad
    //  Serial.println("Sending to iPod...");
    RFduinoBLE.sendByte(WG_Temp);

  while (manualMode == false){        // Only control the temperature when NOT in manual mode.
     /*   During initial turn on, the incoming "hot" water will probably be cold so we need
     *   to purge the cold water.
     
    if (WG_Temp < (iPod_Temp-20)){     // Set valve to full "hot" if water is less than 
                                       // "X" degrees of desired temp. 
      if (!purgingValve) {             // Check to see if we are already purging the valve.
        // Rotate the valve to full "HOT"
        Serial.println("**** Purging 'HOT' water ****");
        Serial.println();
        if (clockwiseHeat == true) {
          myStepper.step(fullOnStep); // Rotate valve 180 degrees (100 steps).  Once we  
                                    // implement the valve positioning reed switch we 
                                    // will rotate the valve till the reed switch activates. 
        }                           // We still want to limit the number of steps to 180 
        else {                      // just in case the reed switch fails. 
          myStepper.step(-fullOnStep); // This will prevent the sketch from hanging.
        }
        purgingValve = true;
        tempReadCount = 0;
      }
    }
      // Purge valve till water temperature attains the desired temp.  Try not to overshoot.
      while (purgingValve == true){
        Serial.println("**** Going through purging loop ****");
        Serial.printf("iPod Temperature is: %d\n",iPod_Temp);
        RFduinoBLE.sendByte(WG_Temp);
        Serial.printf("WG_Temperature is:   %d\n",WG_Temp);
        Serial.println();
        trendTemp2 = trendTemp1;      // Start collecting trend data.  We want the water to 
        trendTemp1 = WG_Temp;         // stabelize close to the desired temperature before 
        tempReadCount += 1;           // we exit the purge function.
  
  //      Serial.printf("Temp2= %d, Temp1= %d, WG_Temp= %d\n",trendTemp2, trendTemp1, WG_Temp);
  //      Serial.println();
       
        if (tempReadCount >= 2) {     // get three water temp samples.
          if (trendTemp2 < trendTemp1 < WG_Temp) {  // determine the water temperature trend.
            trendingUp = true;
          }
          else {
            trendingUp = false;
          }
            trendAmount = (WG_Temp-trendTemp2);  // How much did the temperature change in the 
        }                                        // last 1.5 sec?
        
        if ((WG_Temp >= (iPod_Temp+2)) || (WG_Temp >= absoluteMaxTemp)) { 
        Serial.println("iPod Temp exceeded... Making colder");
          
                                        // Did we overshoot the desired temperature by  
          if (clockwiseHeat == true) {  // more than 2 degrees F? This could happen if 
            myStepper.step(-10);        // the incommnigwater is REALLY hot. If so, get
                                        // it under control now! Rotate valve 18 degrees 
                                        // (10 steps X 1.8   degrees each)COOLER to slow down 
          }                             // the heating.  These are valve rotational degrees 
          else {                        // of the stepper motor which are 1.8 degrees/step, 
            myStepper.step(10);         // 200 steps in a 360 degree full circle.
          }                             //  These are NOT TEMPERATURE DEGREES.
        }
        
        if (((iPod_Temp-5) <= WG_Temp)&&(tempReadCount >= 2)) {  // Are we within 5 
        Serial.println("Getting close to desired temperature");
        Serial.printf("trendAmount = %d\nl", trendAmount);
                                                    // deg F below the desired temperature and 
          if (trendingUp && (trendAmount >= 10)) {   // went through the loop twice? Water heating
             if (clockwiseHeat == true) {                   // fast. Turn down the valve a lot.
              myStepper.step(-bigpurgeStep);      // Rotate valve 18 degrees (10 steps X 1.8     
             }                                    // degrees each) to make it colder.
             else {
              myStepper.step(bigpurgeStep);
             } 
          }                                                         
          else {                                    // If so, slowly start turning the valve
             if (clockwiseHeat == true) {                   // cooler.
              myStepper.step(-purgeStep);                 // Rotate valve 9 degrees (5 steps X 1.8    
             }                                    // degrees each) to make it colder.
             else {
              myStepper.step(purgeStep);
             }      
            }    
          Serial.println("**** Finished Purging Hot Water *****");
          purgingValve = false;
          tempReadCount = 0;        
          break;
        }
        GetWGTemperature();
      } 
  
    // We're done purging the valve, so now we handle regular temperature control here
  
      if (WG_Temp < (iPod_Temp - 0.5)) {      // +/- half a degree for hysteresis
        Serial.print("**** Making water HOTTER using... ");
  //      Serial.println();
        if (WG_Temp < (iPod_Temp - 5)) {      // Heat the water faster if it is to be a 
        Serial.println("BIG steps ****");
           if (clockwiseHeat == true) {             // lot hotter.
              myStepper.step(bigStep);            // Rotate valve 5.4 degrees (3 steps)     
             }
             else {
              myStepper.step(-bigStep);
             }         
        }
        else {
             if (clockwiseHeat == true) {      
        Serial.println("little steps ****");
              myStepper.step(littleStep);             // Rotate valve 1.8 degrees (1 step)     
             }
             else {
              myStepper.step(-littleStep);
             } 
        }
      }
      if (WG_Temp > (iPod_Temp + 0.5)) {      // +/- half a degree for hysteresis
        Serial.print("**** Making water COLDER using...");
  //      Serial.println();
        if (WG_Temp > (iPod_Temp + 5)) {      // Cool the water faster if it is a lot hotter
          if (clockwiseHeat == true) {      
    Serial.println("BIG steps ****");
            myStepper.step(-bigStep);           // Rotate valve 5.4 degrees (3 steps)     
          }
          else {
            myStepper.step(bigStep);
          }         
        }
       else {
         if (clockwiseHeat == true) {      
        Serial.println("little steps ****");
                myStepper.step(-littleStep);           // Rotate valve 1.8 degrees (1 step)     
         }
         else {
          myStepper.step(littleStep);
         } 
       }
     }
  
    if (updateFlash == true){
      SaveToFlash();
    }
   
  }  /********************** End of NORMAL mode of operation ***********************/

  /***********************************************************************
   *                  MANUAL PROGRAMMING MODE...
   *  When the valve is in manual mode, the user adjusts the temperature
   *  by rotating the handle.  The resulting temperature settings are sent
   *  to the user controller and stored as a profile. This process might
   *  be completely transparent to this valve code because all the work 
   *  should done at the user controller end.
   *  We will store the current temperature and count the number of times
   *  that we go through the main loop (this is our time reference) before 
   *  the temperature is changed again.
   **********************************************************************
        //  THIS CODE WILL EVENTUALLY BE MOVED TO THE UI CONTROLLER
        // NOTE:  This code not tested yet...
  if ((manualMode == true)&&(programMode == true)){ // We're wanting to save a user profile.
                                            // Start recording the water temperature & time.
                                            // We will store one temperature and one time duration
                                            //    (in seconds) every three times through the main loop.

      // Wait for water temperature to be stable for about a second.  We don't want to save
      //   temperatures while the user is turing the valve to adjust the temperature.

uint8_t deltaT = abs(WG_Temp - progTemp);   // Get the temperature difference between temperatures.
    if ((col<25)&&(deltaT >= 1.5)){         // 1.5 degrees for hysterysis.                          
      trendTemp1,trendTemp2 = 0;                 
      while ((WG_Temp != trendTemp1) && (WG_Temp != trendTemp2)){ // Stay here till temperature 
        GetWGTemperature();                 // stabilizes.  
        trendTemp2 = trendTemp1;                 
        trendTemp1 = WG_Temp;          
        delay(500);     
      }
      progTemp = WG_Temp;                    // Temperature stable.  Save this temperature for 
      if (col == 0){                         // Compensate for offset of temp & time info.
        programBuff1 [0][col]= progTemp;     // Save the current water temperature.
        programBuff1 [1][col] = runTimeCounter;  // Save the runtime for the last temperature.
      }
      else {
        programBuff1 [0][col]= progTemp;     // Save the current water temperature.
        programBuff1 [1][col-1] = runTimeCounter; // Save the runtime for the last temperature.
      }
      runTimeCounter = 0;                    // Clear our local storage variables. 
      col++;                                 // Point to the next memory location.
      
      if (col == 25){
      Serial.println(" *********** Programming Done **************");

     // WILL WANT TO WRITE programBuff1 to FLASH TO SAVE.
        programMode = false;                // Start up next time in NOrmal Operation Mode.
        SaveToFlash();
      } 
    }
      // Temperature hasn't changed yet so keep counting the times through the main loop
      runTimeCounter++;                     // Keep track of the total time..
  }

  // Back in main loop...
  
  loopCount ++; 

} // END OF MAIN LOOP

/* *************************************************************************************/
/* ******************  End of Main Code ************************************************/
/* *************************************************************************************

// Get all of our saved variables from flash
void Read_Flash( struct flash_data *flash )
{
  Serial.print("  saved_Temp = ");
  saved_Temp = flash->f_saved_Temp;
  Serial.println(saved_Temp, DEC);
  Serial.print("  absoluteMaxTemp = ");
  absoluteMaxTemp = flash->f_absoluteMaxTemp;
  Serial.println(absoluteMaxTemp, DEC);
  Serial.print("  userMaxTemp = ");
  userMaxTemp = flash->f_userMaxTemp;
  Serial.println(userMaxTemp, DEC);  
  Serial.print("  clockwiseHeat = ");
  clockwiseHeat = flash->f_clockwiseHeat;
  Serial.println(clockwiseHeat,HEX);
  Serial.print("  installMode = ");
  installMode = flash->f_installMode;
  Serial.println(installMode, HEX);
  Serial.print("  initializationDone = ");
  initializationDone = flash->f_initializationDone;
  Serial.println(initializationDone, HEX);  
  Serial.print("  bitResolution = ");
  bitResolution = flash->f_bitResolution;
  Serial.println(bitResolution, HEX);  
}

void SaveToFlash(){

//  Since RFduino has a problem, turn off radio to save to flash...
      RFduinoBLE.end();                           // Don't know how long it takes to turn off
                                                  //    so keep the "While radioActive" test in.
      while (RFduinoBLE.radioActive);   // Wait for the radio to go to inactive state
        rc = flashPageErase(PAGE_FROM_ADDRESS(flash));
        if (rc == 0)
          Serial.println("Success");
        else if (rc == 1)
          Serial.println("Error - the flash page is reserved");
        else if (rc == 2)
          Serial.println("Error - the flash page is used by the sketch");
  
     // THIS IS OUR FLASHWRITEBLOCK CODE.  Save the current variable data to flash.                 
//        while (RFduinoBLE.radioActive);   // Wait for the radio to go to inactive state
        rc = flashWriteBlock(flash, &flash_data_values , sizeof(flash_data_values));
        if (rc == 0)
          Serial.println("Success");
        else if (rc == 1)
          Serial.println("Error - the flash page is reserved");
        else if (rc == 2)
          Serial.println("Error - the flash page is used by the sketch");
      
        Serial.println("The data stored in flash contains: ");
        Read_Flash(flash);                        // Write current values to flash.
        updateFlash = false;

//  Turn radio back on..
      RFduinoBLE.begin();                         // Don't know how long it takes to turn
                                                  //   back on.       
}                              

/*****************************************************************************************
                Sample ADC test code.  Modified to read system voltage on P0.30/AIN6
******************************************************************************************/
/* This file contains the source code for a sample application using SAADC.  \
// We can't use timer 0 because it is used by the soft-device.  Change timer number. */

#define SAMPLES_IN_BUFFER 5
volatile uint8_t state = 1;

static const nrf_drv_timer_t m_timer = NRF_DRV_TIMER_INSTANCE(1);     // Eric - changed from 0 to 1
static nrf_saadc_value_t     m_buffer_pool[2][SAMPLES_IN_BUFFER];
static nrf_ppi_channel_t     m_ppi_channel;
static uint32_t              m_adc_evt_counter;

void timer_handler(nrf_timer_event_t event_type, void * p_context)
{
// Nothing in this funciton!!!!
}

void saadc_sampling_event_init(void)
{
    ret_code_t err_code;

    err_code = nrf_drv_ppi_init();
    APP_ERROR_CHECK(err_code);

    nrf_drv_timer_config_t timer_cfg = NRF_DRV_TIMER_DEFAULT_CONFIG;
    timer_cfg.bit_width = NRF_TIMER_BIT_WIDTH_32;
    err_code = nrf_drv_timer_init(&m_timer, &timer_cfg, timer_handler);
    APP_ERROR_CHECK(err_code);

//    setup m_timer for compare event every 400ms 
    uint32_t ticks = nrf_drv_timer_ms_to_ticks(&m_timer, 400);
    nrf_drv_timer_extended_compare(&m_timer,
                                   NRF_TIMER_CC_CHANNEL0,
                                   ticks,
                                   NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK,
                                   false);
    nrf_drv_timer_enable(&m_timer);

    uint32_t timer_compare_event_addr = nrf_drv_timer_compare_event_address_get(&m_timer,
                                                                                NRF_TIMER_CC_CHANNEL0);
    uint32_t saadc_sample_task_addr   = nrf_drv_saadc_sample_task_get();

//    setup ppi channel so that timer compare event is triggering sample task in SAADC 
    err_code = nrf_drv_ppi_channel_alloc(&m_ppi_channel);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_ppi_channel_assign(m_ppi_channel,
                                          timer_compare_event_addr,
                                          saadc_sample_task_addr);
    APP_ERROR_CHECK(err_code);
}

void saadc_sampling_event_enable(void)
{
    ret_code_t err_code = nrf_drv_ppi_channel_enable(m_ppi_channel);

    APP_ERROR_CHECK(err_code);
}

void saadc_callback(nrf_drv_saadc_evt_t const * p_event)
{
    if (p_event->type == NRF_DRV_SAADC_EVT_DONE)
    {
        ret_code_t err_code;

        err_code = nrf_drv_saadc_buffer_convert(p_event->data.done.p_buffer, SAMPLES_IN_BUFFER);
        APP_ERROR_CHECK(err_code);

        int i;
        NRF_LOG_INFO("ADC event number: %d", (int)m_adc_evt_counter);

        for (i = 0; i < SAMPLES_IN_BUFFER; i++)
        {
            NRF_LOG_INFO("%d", p_event->data.done.p_buffer[i]);
        }
        m_adc_evt_counter++;
    }
}

void saadc_init(void)
{
    ret_code_t err_code;
    nrf_saadc_channel_config_t channel_config =
        NRF_DRV_SAADC_DEFAULT_CHANNEL_CONFIG_SE(NRF_SAADC_INPUT_AIN6);      // PS uses AIN6 for Sys_Pwr monitoring

    err_code = nrf_drv_saadc_init(NULL, saadc_callback);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_saadc_channel_init(0, &channel_config);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[0], SAMPLES_IN_BUFFER);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[1], SAMPLES_IN_BUFFER);
    APP_ERROR_CHECK(err_code);
}

// @brief Function for main application entry.
 
uint32_t get_sys_v(void)     // Get the system (Generator) voltage. Sys_V=1.0v for 18V Gen voltage (nominal), 3.06V for 51V Gen voltage.
{                                       
//********************************************************************
// Moved this code into the main function - DELETE WHEN READY
//********************************************************************
/*
    uint32_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();

    ret_code_t ret_code = nrf_pwr_mgmt_init();
    APP_ERROR_CHECK(ret_code);

    saadc_init();
    saadc_sampling_event_init();
*/
//********************************************************************

// We only want to sample when we need the voltage info... Not continually

  int gen_voltage;      // System voltage is derrived from the generator output voltage, nominally 18V.

    saadc_sampling_event_enable();
    NRF_LOG_INFO("SAADC HAL simple example started.");

//    while (1)
//    {
        nrf_pwr_mgmt_run();
        NRF_LOG_FLUSH();
//    }
  return gen_voltage;
}

//********************************************************************
//                        End SAADC code
//********************************************************************


/*****************************************************************************************
                Flash Write Example code to modify for PS
******************************************************************************************/

// @brief This file contains the source code for a sample application using the Flash Write Application.
/* 

#define FLASHWRITE_EXAMPLE_MAX_STRING_LEN       (62u)
#define FLASHWRITE_EXAMPLE_BLOCK_VALID          (0xA55A5AA5)
#define FLASHWRITE_EXAMPLE_BLOCK_INVALID        (0xA55A0000)
#define FLASHWRITE_EXAMPLE_BLOCK_NOT_INIT       (0xFFFFFFFF)

NRF_CLI_UART_DEF(m_cli_uart_transport, 0, 64, 16);
NRF_CLI_DEF(m_cli_uart, "uart_cli:~$ ", &m_cli_uart_transport.transport, '\r', 4);
NRF_MODULE_ENABLED(NRF_CLI)

typedef struct
{
   uint32_t magic_number;
   uint32_t buffer[FLASHWRITE_EXAMPLE_MAX_STRING_LEN + 1]; // + 1 for end of string
} flashwrite_example_flash_data_t;

typedef struct
{
    uint32_t addr;
    uint32_t pg_size;
    uint32_t pg_num;
    flashwrite_example_flash_data_t * m_p_flash_data;
} flashwrite_example_data_t;

static flashwrite_example_data_t m_data;

static ret_code_t clock_config(void)
{
    ret_code_t err_code;

    err_code = nrf_drv_clock_init();
    if (err_code != NRF_SUCCESS && err_code != NRF_ERROR_MODULE_ALREADY_INITIALIZED)
    {
        return err_code;
    }

    nrf_drv_clock_lfclk_request(NULL);

    return NRF_SUCCESS;
}

static void flash_page_init(void)
{
    m_data.pg_num = NRF_FICR->CODESIZE - 1;
    m_data.pg_size = NRF_FICR->CODEPAGESIZE;
    m_data.addr = (m_data.pg_num * m_data.pg_size);

    m_data.m_p_flash_data = (flashwrite_example_flash_data_t *)m_data.addr;

    while (1)
    {
        if (m_data.m_p_flash_data->magic_number == FLASHWRITE_EXAMPLE_BLOCK_VALID)
        {
            return;
        }

        if (m_data.m_p_flash_data->magic_number == FLASHWRITE_EXAMPLE_BLOCK_INVALID)
        {
            ++m_data.m_p_flash_data;
            continue;
        }

        nrf_nvmc_page_erase(m_data.addr);
        return;
    }
}

// @brief Function for application main entry.
 
int write_flash(void)
{
    uint32_t err_code;

    APP_ERROR_CHECK(NRF_LOG_INIT(NULL));

    err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    err_code = clock_config();
    APP_ERROR_CHECK(err_code);

    flash_page_init();

    nrf_drv_uart_config_t uart_config = NRF_DRV_UART_DEFAULT_CONFIG;
    uart_config.pseltxd = TX_PIN_NUMBER;
    uart_config.pselrxd = RX_PIN_NUMBER;
    uart_config.hwfc    = NRF_UART_HWFC_DISABLED;
    err_code = nrf_cli_init(&m_cli_uart, &uart_config, true, true, NRF_LOG_SEVERITY_INFO);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_cli_start(&m_cli_uart);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_RAW_INFO("Flashwrite example started.\r\n");
    NRF_LOG_RAW_INFO("Execute: <flash -h> for more information "
                     "or press the Tab button to see all available commands.\r\n");

    while (true)
    {
        UNUSED_RETURN_VALUE(NRF_LOG_PROCESS());
        nrf_cli_process(&m_cli_uart);
    }
}

static void flash_string_write(uint32_t address, const char * src, uint32_t num_words)
{
    uint32_t i;

    // Enable write.
    NRF_NVMC->CONFIG = NVMC_CONFIG_WEN_Wen;
    while (NRF_NVMC->READY == NVMC_READY_READY_Busy)
    {
    }

    for (i = 0; i < num_words; i++)
    {
        /* Only full 32-bit words can be written to Flash. 
        ((uint32_t*)address)[i] = 0x000000FFUL & (uint32_t)((uint8_t)src[i]);
        while (NRF_NVMC->READY == NVMC_READY_READY_Busy)
        {
        }
    }

    NRF_NVMC->CONFIG = NVMC_CONFIG_WEN_Ren;
    while (NRF_NVMC->READY == NVMC_READY_READY_Busy)
    {
    }
}


static void flashwrite_erase_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    nrf_nvmc_page_erase(m_data.addr);

    m_data.m_p_flash_data = (flashwrite_example_flash_data_t *)m_data.addr;
}

static void flashwrite_read_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    flashwrite_example_flash_data_t * p_data = (flashwrite_example_flash_data_t *)m_data.addr;
    char string_buff[FLASHWRITE_EXAMPLE_MAX_STRING_LEN + 1]; // + 1 for end of string

    if ((p_data == m_data.m_p_flash_data) &&
        (p_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_VALID))
    {
        nrf_cli_fprintf(p_cli, NRF_CLI_WARNING, "Please write something first.\r\n");
        return;
    }

    while (p_data <= m_data.m_p_flash_data)
    {
        if ((p_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_VALID) &&
            (p_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_INVALID))
        {
            nrf_cli_fprintf(p_cli, NRF_CLI_WARNING, "Corrupted data found.\r\n");
            return;
        }
        uint8_t i;
        for (i = 0 ; i <= FLASHWRITE_EXAMPLE_MAX_STRING_LEN; i++)
        {
            string_buff[i] = (char)p_data->buffer[i];
        }

        nrf_cli_fprintf(p_cli, NRF_CLI_NORMAL, "%s\r\n", string_buff);
        ++p_data;
    }
}

static void flashwrite_write_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    static uint16_t const page_size = 4096;

    if (argc < 2)
    {
        nrf_cli_fprintf(p_cli, NRF_CLI_ERROR, "%s:%s", argv[0], " bad parameter count\r\n");
        return;
    }
    if (argc > 2)
    {
        nrf_cli_fprintf(p_cli,
                        NRF_CLI_WARNING,
                        "%s:%s",
                        argv[0],
                        " bad parameter count - please use quotes\r\n");
        return;
    }

    uint32_t len = strlen(argv[1]);
    if (len > FLASHWRITE_EXAMPLE_MAX_STRING_LEN)
    {
        nrf_cli_fprintf(p_cli,
                        NRF_CLI_ERROR,
                        "Too long string. Please limit entered string to %d chars.\r\n",
                        FLASHWRITE_EXAMPLE_MAX_STRING_LEN);
        return;
    }

    if ((m_data.m_p_flash_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_NOT_INIT) &&
        (m_data.m_p_flash_data->magic_number != FLASHWRITE_EXAMPLE_BLOCK_VALID))
    {
        nrf_cli_fprintf(p_cli, NRF_CLI_ERROR, "Flash corrupted, please errase it first.");
        return;
    }

    if (m_data.m_p_flash_data->magic_number == FLASHWRITE_EXAMPLE_BLOCK_VALID)
    {
        uint32_t new_end_addr = (uint32_t)(m_data.m_p_flash_data + 2);
        uint32_t diff = new_end_addr - m_data.addr;
        if (diff > page_size)
        {
            nrf_cli_fprintf(p_cli,
                            NRF_CLI_WARNING,
                            "Not enough space - please erase flash first.\r\n");
            return;
        }
        nrf_nvmc_write_word((uint32_t)&m_data.m_p_flash_data->magic_number,
                            FLASHWRITE_EXAMPLE_BLOCK_INVALID);
        ++m_data.m_p_flash_data;
    }

    //++len -> store also end of string '\0'
    flash_string_write((uint32_t)m_data.m_p_flash_data->buffer, argv[1], ++len);
    nrf_nvmc_write_word((uint32_t)&m_data.m_p_flash_data->magic_number,
                        FLASHWRITE_EXAMPLE_BLOCK_VALID);
}

static void flashwrite_cmd(nrf_cli_t const * p_cli, size_t argc, char **argv)
{
    ASSERT(p_cli);
    ASSERT(p_cli->p_ctx && p_cli->p_iface && p_cli->p_name);

    if ((argc == 1) || nrf_cli_help_requested(p_cli))
    {
        nrf_cli_help_print(p_cli, NULL, 0);
        return;
    }

    nrf_cli_fprintf(p_cli, NRF_CLI_ERROR, "%s:%s%s\r\n", argv[0], " unknown parameter: ", argv[1]);
}

NRF_CLI_CREATE_STATIC_SUBCMD_SET(m_sub_flash)
{
    NRF_CLI_CMD(erase, NULL, "Erase flash.",          flashwrite_erase_cmd),
    NRF_CLI_CMD(read,  NULL, "Read data from flash.", flashwrite_read_cmd),
    NRF_CLI_CMD(write, NULL, "Write data to flash.\n"
                             "Limitations:\n"
                             "- maximum 16 entries,\n"
                             "- each entry is maximum 62 chars long.",
                                                      flashwrite_write_cmd),
    NRF_CLI_SUBCMD_SET_END
};
NRF_CLI_CMD_REGISTER(flash, &m_sub_flash, "Flash access command.", flashwrite_cmd);
*/
/*****************************************************************************************
                End - Flash Write Example code to modify for PS
******************************************************************************************

/*****************************************************************************************
                Set up a millisecond timer for our general use
******************************************************************************************/

void timer_timeout_handler(void * p_context)
{
// put our code here... What happens when the timer times out?
// Sample code - Blink 1 led
  bsp_board_led_on(0);
  nrf_delay_ms(200);   //
  bsp_board_led_off(0);
}

/**@brief Function for starting timers.
 */
void application_timers_start(void)
{
    /* YOUR_JOB: Start your timers. below is an example of how to start a timer.*/
       ret_code_t err_code;
       err_code = app_timer_start(sys_timer, APP_TIMER_TICKS(1000), NULL); //1000 ms= 1 seg
       APP_ERROR_CHECK(err_code);
}

// @brief Function for the Timer initialization.
// @details Initializes the timer module. This creates application timers.
/*
void timers_init(void)      // Eric - This function seems to be a duplicate of "timer_init(void)" 
{
    // Create timers.

    /* YOUR_JOB: Create any timers to be used by the application.
                 Below is an example of how to create a timer.
                 For every new timer needed, increase the value of the macro APP_TIMER_MAX_TIMERS *** Can't find this variable *** by
                 one.
 
       err_code = app_timer_create(&sys_timer, APP_TIMER_MODE_REPEATED, timer_timeout_handler);
       APP_ERROR_CHECK(err_code);
}
/*******************************************************************
/*********** Alternative timer code ********************************
/*******************************************************************

#define LED_TOGGLE_INTERVAL 2000
#define MY_LED   BSP_BOARD_LED_1
static uint32_t led_idx = MY_LED;

static void timers_init(void)
{
    ret_code_t err_code = app_timer_init();           //Initializes the library
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&led_toggle_timer_id, 
                                APP_TIMER_MODE_REPEATED,
                                led_toggle_timeout_handler);
}

void led_toggle_timeout_handler(void *led_idx)
{
    uint8_t *p_led_idx = led_idx;
    bsp_board_led_invert(*p_led_idx);         //Or however you want to turn it on/off
    APP_ERROR_CHECK(err_code);
}

int main (void)
{
  ret_code_t err_code;
  err_code = app_timer_start(led_toggle_timer_id, LED_TOGGLE_INTERVAL, &led_idx);
  APP_ERROR_CHECK(err_code);
}

/*******************************************************************
/*********** Alternative timer code ********************************
/*******************************************************************

void timer_a_handler(void * p_context)
{
    nrf_drv_gpiote_out_set(Kontrol_1);
}
void create_timers()
{   
    uint32_t err_code;
    // Create timers
    err_code = app_timer_create(&m_led_a_timer_id,
                                APP_TIMER_MODE_REPEATED,
                                timer_a_handler);
    APP_ERROR_CHECK(err_code);
}
void start_timer()
{
    uint32_t err_code;

    // Start timer
    err_code = app_timer_start(m_led_a_timer_id,
                                APP_TIMER_TICKS(2),
                                NULL);
    APP_ERROR_CHECK(err_code);
}
 
/**@brief Function for application main entry.
 
int main(void)
{
    // Initialize.
    log_init();
    leds_init();
    timers_init();
    buttons_init();
    power_management_init();
    lfclk_request();
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);
    create_timers();
    start_timer();
    ble_stack_init();
    gap_params_init();
    gatt_init();
    services_init();
    advertising_init();
    conn_params_init();
    sistem_baslangici();
    // Start execution.
    NRF_LOG_INFO("Blinky example started.");
    advertising_start();

    // Enter main loop.
    for (;;)
    {
    //    basla();
         sutun();
       // idle_state_handle();
    }
}

/*******************************************************************
/*********** Alternative timer code ********************************
/*******************************************************************
const nrf_drv_timer_t sys_timer = NRF_DRV_TIMER_INSTANCE(1);      // use timer 1 for non-softdevice timers was 0


// * @brief Handler for timer events.
// * This routine uses leds to show what's happening we'll change this for our use later
 
void timer_led_event_handler(nrf_timer_event_t event_type, void* p_context)
{
    static uint32_t i;
    uint32_t led_to_invert = ((i++) % LEDS_NUMBER);

    switch (event_type)
    {
        case NRF_TIMER_EVENT_COMPARE1:              // timer 1 was 0
            bsp_board_led_invert(led_to_invert);
            break;

        default:
            //Do nothing.
            break;
    }
}

// @brief Function for main application entry.
 
int ticks_timer(void)
{
    uint32_t time_ms = 20; //Time(in miliseconds) between consecutive compare events (20ms for stepper motor speed)
    uint32_t time_ticks;
    uint32_t err_code = NRF_SUCCESS;

    //Configure all leds on board.
    bsp_board_init(BSP_INIT_LEDS);   // Already set up above.

    //Configure sys_timer for generating simple light effect - leds on board will invert his state one after the other.
    nrf_drv_timer_config_t timer_cfg = NRF_DRV_TIMER_DEFAULT_CONFIG;
    err_code = nrf_drv_timer_init(&sys_timer, &timer_cfg, timer_led_event_handler);
    APP_ERROR_CHECK(err_code);

    time_ticks = nrf_drv_timer_ms_to_ticks(&sys_timer, time_ms);

    nrf_drv_timer_extended_compare(
         &sys_timer, NRF_TIMER_CC_CHANNEL1, time_ticks, NRF_TIMER_SHORT_COMPARE1_CLEAR_MASK, true);

    nrf_drv_timer_enable(&sys_timer);
}
/*****************************************************************************************
                  END - Set up a millisecond timer for our general use
******************************************************************************************/


/*****************************************************************************************
                          ALTERNATIVE FLASH CODE
******************************************************************************************/
/*

static void fstorage_test(void)
{
		static uint32_t data;
		uint32_t flash_data[4];
	
		FS_REGISTER_CFG(fs_config_t fs_config) =
		{
			  .callback  = fs_evt_handler, // Function for event callbacks.
				.num_pages = NUM_PAGES,      // Number of physical flash pages required.
				.priority  = 0xFE            // Priority for flash usage.
		};
		
		fs_ret_t ret = fs_init();
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		
		// Erase one page (page 0).
		printf("Erasing a flash page at address 0x%X\r\n", (uint32_t)fs_config.p_start_addr);
		fs_callback_flag = 1;
		ret = fs_erase(&fs_config, fs_config.p_start_addr, 1);
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		while(fs_callback_flag == 1)  { power_manage(); }
		
		//Read the first 4 words of the page
		printf("Data read from flash address 0x%X: ", (uint32_t)fs_config.p_start_addr);
		for(int i=0; i<4; i++)
		{
				flash_data[i] = *(fs_config.p_start_addr + i);
				printf("%X ", flash_data[i]);
		}
		printf("\r\n");
		
		//Write the first byte
		data = 0xAAAAAAAA;
		printf("Writing data 0x%X to address 0x%X\r\n", data, (uint32_t)fs_config.p_start_addr);
		fs_callback_flag = 1;
		ret = fs_store(&fs_config, fs_config.p_start_addr, &data, 1);      //Write data to memory address 0x0003F000. Check it with command: nrfjprog --memrd 0x0003F000 --n 16
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		while(fs_callback_flag == 1)  { power_manage(); }
		
		//Write second byte
		data = 0xBBBBBBBB;
		printf("Writing data 0x%X to address 0x%X\r\n", data, (uint32_t)fs_config.p_start_addr + 4);
		fs_callback_flag = 1;
		ret = fs_store(&fs_config, fs_config.p_start_addr + 1, &data, 1);      //Write data to memory address 0x0003F000. Check it with command: nrfjprog --memrd 0x0003F000 --n 16
		if (ret != FS_SUCCESS)
		{
				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
		}
		while(fs_callback_flag == 1)  { power_manage(); }
		
		//Read the first 4 words of the page
		printf("Data read from flash address 0x%X: ", (uint32_t)fs_config.p_start_addr);
		for(int i=0; i<4; i++)
		{
				flash_data[i] = *(fs_config.p_start_addr + i);
				printf("%X ", flash_data[i]);
		}
		printf("\r\n");
}


/*****************************************************************************************
                        END - ALTERNATIVE FLASH CODE
******************************************************************************************/
/*
/*****************************************************************************************
                        ANOTHER FLASH WRITE EXAMPLE
******************************************************************************************/
/*
fds_init();
nrf_delay_ms(300);

__ALIGN(4) beacon_packet_struct   beacon_flags;

typedef struct
{
    uint8_t led_blink; //was bool
    uint8_t initial_adv_disable; //was bool
    uint8_t enable_ibeacon_Packet; //was bool
    uint8_t enable_cbeacon1_Health_Packet; //was bool
    uint8_t enable_cbeacon2_Proximity_Packet;  //was bool
    uint8_t enable_ibeacon_SOS_Packet; //was bool
    uint8_t enable_Mems; //was bool
}beacon_packet_struct;


int fds_write_struct(uint16_t file_id, uint16_t flash_key)
{
    ret_code_t ret;

    fds_record_t        record;
    fds_record_desc_t   record_desc;

    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

	if (flash_key == BEACON_STATE_KEY)
    {

        record.file_id              = FILE_ID;
        record.key                  = BEACON_STATE_KEY;
        record.data.p_data          = &beacon_flags;  //&beacon_flags;
        record.data.length_words    = ( sizeof(beacon_packet_struct) + 3 ) / sizeof(uint32_t);
      
        ret_code_t rc = fds_record_find(FILE_ID, BEACON_STATE_KEY , &desc, &tok);  //Search for record. Update if found else write a new one

        if (rc == FDS_SUCCESS)
        {   
            NRF_LOG_INFO("Similar BEACON STATE Variables Record was found thus it is being updated");
            rc = fds_record_update(&desc, &record);
            APP_ERROR_CHECK(rc);
        }
        else
        {
            ret_code_t ret = fds_record_write(&record_desc, &record);
            APP_ERROR_CHECK(ret);
            NRF_LOG_INFO("First time writing BEACON STATE record");
        }

        return 1;
    }
	
}
/*****************************************************************************************
                      END- ANOTHER FLASH WRITE EXAMPLE
******************************************************************************************/
/*
/*****************************************************************************************
                  MORE ALTERNATIVE FLASH CODE
/******************************************************************************************/
/*
#include <stdint.h>
#include <string.h>
#include "nrf.h"
#include "nordic_common.h"
#ifdef SOFTDEVICE_PRESENT
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#else
#include "nrf_drv_clock.h"
#endif
#include "fds.h"
#include "app_timer.h"
#include "app_error.h"
#include "nrf_cli.h"
#include "fds_example.h"

#define NRF_LOG_MODULE_NAME app
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
*/

/* A tag identifying the SoftDevice BLE configuration. */
#define APP_BLE_CONN_CFG_TAG    1


/* Array to map FDS return values to strings. */
char const * fds_err_str[] =
{
    "FDS_SUCCESS",
    "FDS_ERR_OPERATION_TIMEOUT",
    "FDS_ERR_NOT_INITIALIZED",
    "FDS_ERR_UNALIGNED_ADDR",
    "FDS_ERR_INVALID_ARG",
    "FDS_ERR_NULL_ARG",
    "FDS_ERR_NO_OPEN_RECORDS",
    "FDS_ERR_NO_SPACE_IN_FLASH",
    "FDS_ERR_NO_SPACE_IN_QUEUES",
    "FDS_ERR_RECORD_TOO_LARGE",
    "FDS_ERR_NOT_FOUND",
    "FDS_ERR_NO_PAGES",
    "FDS_ERR_USER_LIMIT_REACHED",
    "FDS_ERR_CRC_CHECK_FAILED",
    "FDS_ERR_BUSY",
    "FDS_ERR_INTERNAL",
};

/* Array to map FDS events to strings. */
static char const * fds_evt_str[] =
{
    "FDS_EVT_INIT",
    "FDS_EVT_WRITE",
    "FDS_EVT_UPDATE",
    "FDS_EVT_DEL_RECORD",
    "FDS_EVT_DEL_FILE",
    "FDS_EVT_GC",
};

/* Dummy configuration data. */
static configuration_t m_dummy_cfg =
{
    .config1_on  = false,
    .config2_on  = true,
    .boot_count  = 0x0,
    .device_name = "dummy",
};

/* A record containing dummy configuration data. */
static fds_record_t const m_dummy_record =
{
    .file_id           = CONFIG_FILE,
    .key               = CONFIG_REC_KEY,
    .data.p_data       = &m_dummy_cfg,
    /* The length of a record is always expressed in 4-byte units (words). */
    .data.length_words = (sizeof(m_dummy_cfg) + 3) / sizeof(uint32_t),
};

/* Keep track of the progress of a delete_all operation. */
static struct
{
    bool delete_next;   //!< Delete next record.
    bool pending;       //!< Waiting for an fds FDS_EVT_DEL_RECORD event, to delete the next record.
} m_delete_all;

/* Flag to check fds initialization. */
static bool volatile m_fds_initialized;


static void fds_evt_handler(fds_evt_t const * p_evt)
{
    NRF_LOG_GREEN("Event: %s received (%s)",
                  fds_evt_str[p_evt->id],
                  fds_err_str[p_evt->result]);

    switch (p_evt->id)
    {
        case FDS_EVT_INIT:
            if (p_evt->result == FDS_SUCCESS)
            {
                m_fds_initialized = true;
            }
            break;

        case FDS_EVT_WRITE:
        {
            if (p_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_INFO("Record ID:\t0x%04x",  p_evt->write.record_id);
                NRF_LOG_INFO("File ID:\t0x%04x",    p_evt->write.file_id);
                NRF_LOG_INFO("Record key:\t0x%04x", p_evt->write.record_key);
            }
        } break;

        case FDS_EVT_DEL_RECORD:
        {
            if (p_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_INFO("Record ID:\t0x%04x",  p_evt->del.record_id);
                NRF_LOG_INFO("File ID:\t0x%04x",    p_evt->del.file_id);
                NRF_LOG_INFO("Record key:\t0x%04x", p_evt->del.record_key);
            }
            m_delete_all.pending = false;
        } break;

        default:
            break;
    }
}


/**@brief   Begin deleting all records, one by one. */
void delete_all_begin(void)
{
    m_delete_all.delete_next = true;
}


/**@brief   Process a delete all command.
 *
 * Delete records, one by one, until no records are left.
 */
void delete_all_process(void)
{
    if (   m_delete_all.delete_next
        & !m_delete_all.pending)
    {
        NRF_LOG_INFO("Deleting next record.");

        m_delete_all.delete_next = record_delete_next();
        if (!m_delete_all.delete_next)
        {
            NRF_LOG_CYAN("No records left to delete.");
        }
    }
}

/* **************************** THIS CODE ALREADY IMPLEMENTED ******************
#ifdef SOFTDEVICE_PRESENT
/**@brief   Function for initializing the SoftDevice and enabling the BLE stack. 
static void ble_stack_init(void)
{
    ret_code_t rc;
    uint32_t   ram_start;

    /* Enable the SoftDevice. 
    rc = nrf_sdh_enable_request();
    APP_ERROR_CHECK(rc);

    rc = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(rc);

    rc = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(rc);
}
#else
*/
static void clock_init(void)
{
    /* Initialize the clock. */
    ret_code_t rc = nrf_drv_clock_init();
    APP_ERROR_CHECK(rc);

    nrf_drv_clock_lfclk_request(NULL);

    /* Wait for the clock to be ready. */
    while (!nrf_clock_lf_is_running()) {;}
}

/**@brief   Initialize the timer. 
static void timer_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);
}
*/
/*
/**@brief   Initialize logging.
static void log_init(void)
{
    ret_code_t rc = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(rc);
}
*/

/**@brief   Sleep until an event is received. */
static void power_manage(void)
{
#ifdef SOFTDEVICE_PRESENT
    (void) sd_app_evt_wait();
#else
    __WFE();
#endif
}


/**@brief   Wait for fds to initialize. */
static void wait_for_fds_ready(void)
{
    while (!m_fds_initialized)
    {
        power_manage();
    }
}


int access_flash(void)
{
    ret_code_t rc;
/*
#ifdef SOFTDEVICE_PRESENT
    ble_stack_init();
#else
    clock_init();
#endif

    timer_init();
    log_init();
    cli_init();

    NRF_LOG_INFO("FDS example started.")

    /* Register first to receive an event when initialization is complete. */
    (void) fds_register(fds_evt_handler);

    NRF_LOG_INFO("Initializing fds...");

    rc = fds_init();
    APP_ERROR_CHECK(rc);

    /* Wait for fds to initialize. */
    wait_for_fds_ready();

    NRF_LOG_INFO("Available commands:");
    NRF_LOG_INFO("- print all\t\tprint records");
    NRF_LOG_INFO("- print config\tprint configuration");
    NRF_LOG_INFO("- update\t\tupdate configuration");
    NRF_LOG_INFO("- stat\t\tshow statistics");
    NRF_LOG_INFO("- write\t\twrite a new record");
    NRF_LOG_INFO("- delete\t\tdelete a record");
    NRF_LOG_INFO("- delete_all\tdelete all records");
    NRF_LOG_INFO("- gc\t\trun garbage collection");

    NRF_LOG_INFO("Reading flash usage statistics...");

    fds_stat_t stat = {0};

    rc = fds_stat(&stat);
    APP_ERROR_CHECK(rc);

    NRF_LOG_INFO("Found %d valid records.", stat.valid_records);
    NRF_LOG_INFO("Found %d dirty records (ready to be garbage collected).", stat.dirty_records);

    fds_record_desc_t desc = {0};
    fds_find_token_t  tok  = {0};

    rc = fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok);

    if (rc == FDS_SUCCESS)
    {
        /* A config file is in flash. Let's update it. */
        fds_flash_record_t config = {0};

        /* Open the record and read its contents. */
        rc = fds_record_open(&desc, &config);
        APP_ERROR_CHECK(rc);

        /* Copy the configuration from flash into m_dummy_cfg. */
        memcpy(&m_dummy_cfg, config.p_data, sizeof(configuration_t));

        NRF_LOG_INFO("Config file found, updating boot count to %d.", m_dummy_cfg.boot_count);

        /* Update boot count. */
        m_dummy_cfg.boot_count++;

        /* Close the record when done reading. */
        rc = fds_record_close(&desc);
        APP_ERROR_CHECK(rc);

        /* Write the updated record to flash. */
        rc = fds_record_update(&desc, &m_dummy_record);
        APP_ERROR_CHECK(rc);
    }
    else
    {
        /* System config not found; write a new one. */
        NRF_LOG_INFO("Writing config file...");

        rc = fds_record_write(&desc, &m_dummy_record);
        APP_ERROR_CHECK(rc);
    }

    cli_start();

    /* Enter main loop. */
    for (;;)
    {
        if (!NRF_LOG_PROCESS())
        {
            power_manage();
        }
        cli_process();
        delete_all_process();
    }
}
/**
 * @}
 */
/*****************************************************************************************
                END - MORE ALTERNATIVE FLASH CODE
/******************************************************************************************/
